"""Test cases for testing hosting services."""

from __future__ import unicode_literals

import io
import json
from contextlib import contextmanager

from django.utils import six
from django.utils.six.moves.urllib.error import HTTPError
from django.utils.six.moves.urllib.parse import urlparse
from kgb import SpyAgency

from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.hostingsvcs.service import get_hosting_service
from reviewboard.testing import TestCase


class HttpTestContext(object):
    """State and functions for an HTTP test.

    This is provided when calling
    :py:meth:`HostingServiceTestCase.setup_http_test`. It serves up state
    that can be tested with, along with useful functions for creating necessary
    objects and asserting results.

    Version Added:
        3.0.4

    Attributes:
        client (reviewboard.hostingsvcs.service.HostingServiceClient):
            The hosting service client used to perform HTTP requests for
            this test.

        hosting_account (reviewboard.hostingsvcs.models.HostingServiceAccount):
            The hosting account used for the test.

        service (reviewboard.hostingsvcs.service.HostingService):
            The hosting service instance used for the test.
    """

    def __init__(self, test_case, hosting_account, http_request_func):
        """Initialize the test context.

        Args:
            test_case (HostingServiceTestCase):
                The parent test case.

            hosting_account (reviewboard.hostingsvcs.models.
                             HostingServiceAccount):
                The hosting service account set up for the test.

            http_request_func (callable):
                The function used to handle HTTP requests.
        """
        self.hosting_account = hosting_account
        self.service = hosting_account.service
        self.client = self.service.client

        self._test_case = test_case
        self._http_request_func = http_request_func

    @property
    def http_calls(self):
        """The HTTP calls made by the service.

        This is a list of spy calls from KGB.
        """
        return self._http_request_func.calls

    def create_repository(self, **kwargs):
        """Create a repository using the current hosting account.

        This wraps :py:meth:`HostingServiceTestCase.create_repository`,
        specifying the hosting account that was set up in this test context.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to
                :py:meth:`HostingServiceTestCase.create_repository`.

        Returns:
            reviewboard.scmtools.models.Repository:
            The new repository.
        """
        return self._test_case.create_repository(
            hosting_account=self.hosting_account,
            **kwargs)

    def assertHTTPCall(self, index=0, method='GET', body=None, headers=None,
                       **kwargs):
        """Assert that an HTTP call was made.

        This sets some defaults based on the test case, helping both to
        eliminate the amount of data that needs to be checked and helping
        avoid missed data.

        Any keyword argument accepted by
        :py:meth:`HostingServiceClient.http_request()
        <reviewboard.hostingsvcs.service.HostingServiceClient.http_request>`
        can be provided.

        If ``username`` or ``password`` are not explicitly provided, the
        values from :py:attr:`HostingServiceTestCase.default_username` or
        :py:attr:`HostingServiceTestCase.default_password` will be used.

        Args:
            index (int, optional):
                The index of the HTTP call.

            method (unicode, optional);
                The HTTP method expected for the call.

            body (unicode, optional):
                The expected body for the call (used for POST/PUT requests).

            headers (dict, optional):
                The expected headers for the call.

            **kwargs (dict):
                Additional parameters to check in the call.
        """
        kwargs.setdefault('username', self._test_case.default_username)
        kwargs.setdefault('password', self._test_case.default_password)

        self._test_case.assertTrue(self.http_calls[index].called_with(
            method=method,
            body=body,
            headers=headers,
            **kwargs))


class HostingServiceTestCase(SpyAgency, TestCase):
    """Base class for unit tests for hosting services."""

    #: The registered name of the service.
    service_name = None

    #: The default username to use for accounts.
    default_username = 'myuser'

    #: The default password to use for accounts.
    default_password = 'mypass'

    #: The default hosting URL to use for accounts.
    #:
    #: This is only used for accounts set with ``use_url=True`` (or when
    #: ``default_use_hosting_url`` is ``True``).
    default_hosting_url = 'https://example.com'

    #: Whether to create accounts attached to a hosting URL by default.
    default_use_hosting_url = False

    #: Default data to set for created hosting service accounts.
    default_account_data = {}

    #: Default data to set for created repositories.
    default_repository_extra_data = {}

    #: Default SCMTool to set for created repositories.
    default_repository_tool_name = None

    fixtures = ['test_scmtools']

    @classmethod
    def setUpClass(cls):
        super(HostingServiceTestCase, cls).setUpClass()

        if cls.service_name:
            cls.service_class = get_hosting_service(cls.service_name)
        else:
            cls.service_class = None

    def setUp(self):
        super(HostingServiceTestCase, self).setUp()

        self.assertIsNotNone(self.service_class)

    @contextmanager
    def setup_http_test(self, http_request_func=None, payload=None,
                        headers=None, status_code=None, hosting_account=None,
                        expected_http_calls=None):
        """Set up state for HTTP-related tests.

        This takes the hard work out of testing hosting service functionality
        that needs to communicate over HTTP. It's a context manager that takes
        in information on what to return to the client and yields a context
        containing state and functions for performing and checking calls.

        By default, this will cause any HTTP requests to return an empty
        byte string as a payload. Explicit payloads can also be provided,
        as can an HTTP error code. For more advanced needs, a function can
        be provided that will handle the HTTP request (which is useful when
        your test may invoke several HTTP endpoints).

        This may be called repeatedly in the same test function.

        Args:
            http_request_func (callable, optional):
                An explicit HTTP request function to call when performing an
                HTTP request. This will override
                :py:meth:`reviewboard.hostingsvcs.service.HostingServiceClient
                .http_request`.

            payload (bytes, optional):
                An explicit payload to return to the client.

            headers (dict, optional):
                Headers to send along with the result.

            status_code (int, optional):
                An explicit HTTP status code. Only values >= 400 are used.
                Providing an error code will raise an HTTP error to the
                client, using the ``payload`` value if provided.

            hosting_account (reviewboard.hostingsvcs.models.
                             HostingServiceAccount, optional):
                An explicit hosting account to use. If not provided, one will
                be created using :py:meth:`create_hosting_account`.

            expected_http_calls (int, optional):
                The number of HTTP calls expected. If provided, this will
                assert that there were this many HTTP calls.

        Context:
            HttpTestContext:
            The context for the test, containing state, helper functions,
            and results.
        """
        if hosting_account is None:
            hosting_account = self.create_hosting_account()

        if http_request_func:
            # Sanity-check that the caller isn't mixing incompatible arguments.
            if payload is not None:
                raise ValueError(
                    'http_request_func and payload cannot both be provided')
            elif payload is not None:
                raise ValueError(
                    'http_request_func and status_code cannot both be '
                    'provided')
        else:
            http_request_func = self.make_handler_for_paths({
                None: {
                    'status_code': status_code,
                    'payload': payload,
                    'headers': headers,
                },
            })

        client = hosting_account.service.client

        if hasattr(client.http_request, 'unspy'):
            # Reset for this next test. This allows the test case to use
            # this context function multiple times.
            client.http_request.unspy()

        self.spy_on(client.http_request, call_fake=http_request_func)

        ctx = HttpTestContext(self, hosting_account, client.http_request)
        yield ctx

        if expected_http_calls is not None:
            self.assertEqual(len(ctx.http_calls), expected_http_calls)

    def make_handler_for_paths(self, paths):
        """Return an HTTP handler function for serving the supplied paths.

        This is meant to be passed to :py:meth:`setup_http_test`.

        This takes a dictionary matching paths to information to return. Each
        key is a path relative to the domain, which may optionally contain a
        full query string to match. It may also be ``None``, which is the
        fallback.

        Each value is a dictionary containing optional ``payload``,
        ``status_code``, or ``headers`` values.

        Args:
            paths (dict):
                The dictionary of paths.

        Returns:
            callable:
            The resulting HTTP handler function.

        Example:
            .. code-block:: python

               handler = make_handler_for_paths({
                   '/api/1/diffs/': {
                       'payload': b'...',
                       'headers': {
                           b'My-Header': b'value',
                        },
                   },
                   '/api/1/bad/': {
                       'status_code': 404,
                       'payload': b'Not found.',
                   },
                   None: {
                       'payload': b'fallback data...',
                   },
               })
        """
        # Validate the paths to make sure payloads are in the right format.
        for path, path_info in six.iteritems(paths):
            payload = path_info.get('payload')

            if payload is not None and not isinstance(payload, bytes):
                raise TypeError('payload must be a byte string or None')

        def _handler(client, url, *args, **kwargs):
            parts = urlparse(url)

            path_info = paths.get('%s?%s' % (parts.path, parts.query))

            if path_info is None:
                path_info = paths.get(parts.path)

                if path_info is None:
                    path_info = paths.get(None)

                    if path_info is None:
                        self.fail('Unexpected path "%s"' % parts.path)

            status_code = path_info.get('status_code')
            payload = path_info.get('payload') or b''
            headers = path_info.get('headers') or {}

            if status_code is not None and status_code >= 400:
                raise HTTPError(url, status_code, '', headers,
                                io.BytesIO(payload))
            else:
                return payload, headers

        return _handler

    def dump_json(self, data):
        """Dump JSON-compatible data to a byte string.

        Args:
            data (object):
                The data to dump.

        Returns:
            bytes:
            The serialized byte string.
        """
        result = json.dumps(data)

        if isinstance(result, six.text_type):
            result = result.encode('utf-8')

        return result

    def get_form(self, plan=None, fields={}):
        """Return the configuration form for the hosting service.

        Args:
            plan (unicode, optional):
                The hosting plan that the configuration form is for.

            fields (dict, optional):
                The initial field data to populate the form with.

        Returns:
            reviewboard.hostingsvcs.forms.HostingServiceForm:
        """
        form_cls = self.service_class.get_field(name='form', plan=plan)
        self.assertIsNotNone(form_cls)

        form = form_cls(fields)
        self.assertTrue(form.is_valid())

        return form

    def create_hosting_account(self, use_url=None, local_site=None,
                               data=None):
        """Create a hosting account to test with.

        Args:
            use_url (unicode, optional):
                Whether the account should be attached to a given hosting URL,
                for self-hosted services. If set, this will use
                ``https://example.com``.

            local_site (reviewboard.site.models.LocalSite, optional):
                A Local Site to attach the account to.

            data (dict, optional):
                Optional data to set for the account. If this is ``None``,
                :py:attr:`default_account_data` will be used.

        Returns:
            reviewboard.hostingsvcs.models.HostingServiceAccount:
            The new hosting service account.
        """
        if use_url is None:
            use_url = self.default_use_hosting_url

        if use_url:
            hosting_url = self.default_hosting_url
        else:
            hosting_url = None

        account = HostingServiceAccount(service_name=self.service_name,
                                        username=self.default_username,
                                        hosting_url=hosting_url,
                                        local_site=local_site)

        if data is not None:
            account.data = data
        else:
            account.data = self.default_account_data

        account.save()

        return account

    def create_repository(self, **kwargs):
        """Create a repository for a test.

        This wraps :py:meth:`TestCase.create_repository()
        <reviewboard.testing.testcase.TestCase.create_repository>`,
        specifying a default SCM Tool name (if
        :py:attr:`default_repository_tool_name`) is set) and
        extra data (if :py:attr:`default_repository_extra_data` is set).

        Args:
            **kwargs (dict):
                Keyword arguments to pass to
                :py:meth:`TestCase.create_repository()
                <reviewboard.testing.testcase.TestCase.create_repository>`.

        Returns:
            reviewboard.scmtools.models.Repository:
            The new repository.
        """
        if 'extra_data' not in kwargs:
            kwargs['extra_data'] = self.default_repository_extra_data.copy()

        return super(HostingServiceTestCase, self).create_repository(**kwargs)

    def get_repository_fields(self, tool_name, fields, plan=None,
                              with_url=None, hosting_account=None):
        """Return populated fields for a repository.

        Args:
            tool_name (unicode):
                The name of the SCM Tool used for the repository.

            fields (dict):
                A dictionary of fields for the hosting service form.

            plan (unicode, optional):
                The optional hosting plan to use for the configuration.

            with_url (unicode, optional):
                Whether the account should be attached to a given hosting URL,
                for self-hosted services. If set, this will use
                ``https://example.com``. This is ignored if ``hosting_account``
                is provided.

                This value defaults to :py:attr:`default_use_hosting_url`

            hosting_account (reviewboard.hostingsvcs.models.
                             HostingServiceAccount, optional):
                An explicit hosting service account to use.

        Returns:
            dict:
            The populated field data for the repository.
        """
        form = self.get_form(plan, fields)

        if not hosting_account:
            hosting_account = self.create_hosting_account(with_url)

        service = hosting_account.service
        self.assertIsNotNone(service)

        field_vars = form.clean().copy()
        field_vars.update(hosting_account.data)

        return service.get_repository_fields(
            username=hosting_account.username,
            hosting_url=self.default_hosting_url,
            plan=plan,
            tool_name=tool_name,
            field_vars=field_vars)



"""Unit tests for reviewboard.accounts.forms.pages.PrivacyForm."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.core.cache import cache
from django.test.client import RequestFactory
from djblets.privacy.consent import (get_consent_requirements_registry,
                                     get_consent_tracker)
from djblets.privacy.consent.common import PolicyConsentRequirement

from reviewboard.accounts.forms.pages import PrivacyForm
from reviewboard.accounts.pages import AccountPage
from reviewboard.accounts.views import MyAccountView
from reviewboard.testing import TestCase


class MyAccountViewTests(TestCase):
    """Unit tests for MyAccountView."""

    fixtures = ['test_users']

    def tearDown(self):
        super(MyAccountViewTests, self).tearDown()

        cache.clear()

    def test_render_all_accept_requirements(self):
        """Testing MyAccountView renders all forms when a user has accepted all
        requirements
        """
        settings = {
            'privacy_enable_user_consent': True,
        }
        user = User.objects.get(username='doc')
        get_consent_tracker().record_consent_data_list(
            user,
            [
                requirement.build_consent_data(granted=True)
                for requirement in get_consent_requirements_registry()
            ])

        request = RequestFactory().get('/account/preferences')
        request.user = User.objects.get(username='doc')
        view = MyAccountView()

        self.client.login(username='doc', password='doc')

        with self.siteconfig_settings(settings):
            rsp = self.client.get('/account/preferences/')

            self.assertEqual(rsp.status_code, 200)
            context = rsp.context

            self.assertEqual(context['render_sidebar'], True)
            self.assertEqual(
                {
                    type(page)
                    for page in context['pages']
                    if page.is_visible()
                },
                {
                    account_page
                    for account_page in AccountPage.registry
                    if account_page(view, request, request.user).is_visible()
                })

    def test_render_all_reject_requirements(self):
        """Testing MyAccountView renders all forms when a user has rejected all
        consent decisions
        """
        settings = {
            'privacy_enable_user_consent': True,
        }
        user = User.objects.get(username='doc')
        get_consent_tracker().record_consent_data_list(
            user,
            [
                requirement.build_consent_data(granted=False)
                for requirement in get_consent_requirements_registry()
            ])

        request = RequestFactory().get('/account/preferences')
        request.user = User.objects.get(username='doc')
        view = MyAccountView()

        self.client.login(username='doc', password='doc')

        with self.siteconfig_settings(settings):
            rsp = self.client.get('/account/preferences/')

            self.assertEqual(rsp.status_code, 200)
            context = rsp.context

            self.assertEqual(context['render_sidebar'], True)
            self.assertEqual(
                {
                    type(page)
                    for page in context['pages']
                },
                {
                    account_page
                    for account_page in AccountPage.registry
                    if account_page(view, request, request.user).is_visible()
                })

    def test_render_only_privacy_form_if_missing_consent(self):
        """Testing MyAccountView only renders privacy form when a user has
        pending consent decisions
        """
        settings = {
            'privacy_enable_user_consent': True,
        }

        self.client.login(username='doc', password='doc')

        with self.siteconfig_settings(settings):
            rsp = self.client.get('/account/preferences/')

        self.assertEqual(rsp.status_code, 200)
        context = rsp.context

        self.assertEqual(context['render_sidebar'], False)
        self.assertEqual(len(context['forms']), 1)
        self.assertIsInstance(context['forms'][0], PrivacyForm)

    def test_render_only_privacy_form_if_reject_policy_grant_others(self):
        """Testing MyAccountView only renders privacy policy when a user has
        rejected the privacy policy/terms of service and granted all other
        requirements
        """
        settings = {
            'privacy_enable_user_consent': True,
            'privacy_policy_url': 'https://example.com',
            'terms_of_service_url': 'https://example.com',
        }

        user = User.objects.get(username='doc')

        # Accept all consent requirements *except* the policy.
        get_consent_tracker().record_consent_data_list(
            user,
            [
                requirement.build_consent_data(
                    granted=not isinstance(requirement,
                                           PolicyConsentRequirement))
                for requirement in get_consent_requirements_registry()
            ])

        self.client.login(username='doc', password='doc')

        with self.siteconfig_settings(settings):
            rsp = self.client.get('/account/preferences/')

        self.assertEqual(rsp.status_code, 200)
        context = rsp.context

        self.assertEqual(context['render_sidebar'], False)
        self.assertEqual(len(context['forms']), 1)
        self.assertIsInstance(context['forms'][0], PrivacyForm)

    def test_render_only_privacy_form_if_reject_policy_reject_others(self):
        """Testing MyAccountView only renders privacy policy when a user has
        rejected the privacy policy/terms of service and rejected all other
        requirements
        """
        settings = {
            'privacy_enable_user_consent': True,
            'privacy_policy_url': 'https://example.com',
            'terms_of_service_url': 'https://example.com',
        }

        user = User.objects.get(username='doc')

        # Accept all consent requirements *except* the policy.
        get_consent_tracker().record_consent_data_list(
            user,
            [
                requirement.build_consent_data(granted=False)
                for requirement in get_consent_requirements_registry()
            ])

        self.client.login(username='doc', password='doc')

        with self.siteconfig_settings(settings):
            rsp = self.client.get('/account/preferences/')

        self.assertEqual(rsp.status_code, 200)
        context = rsp.context

        self.assertEqual(context['render_sidebar'], False)
        self.assertEqual(len(context['forms']), 1)
        self.assertIsInstance(context['forms'][0], PrivacyForm)

    def test_redirect_privacy_form(self):
        """Testing MyAccountView redirects to previous URL when saving the
        privacy form if a next URL is provided
        """
        settings = {
            'privacy_enable_user_consent': True,
        }

        self.client.login(username='doc', password='doc')

        with self.siteconfig_settings(settings):
            rsp = self.client.post(
                '/account/preferences/',
                dict({
                    'next_url': '/some-page/',
                    'form_target': PrivacyForm.form_id,
                }, **{
                    'consent_%s_choice' % requirement.requirement_id: 'allow'
                    for requirement in get_consent_requirements_registry()

                }))

        self.assertEqual(rsp.status_code, 302)
        self.assertEqual(rsp.url, 'http://testserver/some-page/')



from __future__ import unicode_literals

import logging

from django.contrib.auth.models import User
from django.template import Context, RequestContext, Template
from django.test import RequestFactory
from djblets.testing.decorators import add_fixtures
from kgb import SpyAgency

from reviewboard.accounts.trophies import TrophyType, trophies_registry
from reviewboard.reviews.fields import (BaseReviewRequestField,
                                        BaseReviewRequestFieldSet,
                                        register_review_request_fieldset,
                                        unregister_review_request_fieldset)
from reviewboard.reviews.models import Comment
from reviewboard.testing import TestCase


class DisplayReviewRequestTrophiesTests(TestCase):
    fixtures = ['test_users']

    @classmethod
    def setUpClass(cls):
        super(DisplayReviewRequestTrophiesTests, cls).setUpClass()

        cls._request_factory = RequestFactory()

    def tearDown(self):
        super(DisplayReviewRequestTrophiesTests, self).tearDown()

        trophies_registry.reset()

    def test_old_style_trophy(self):
        """Testing {% display_review_request_trophies %} for old-style
        TrophyType
        """
        class OldTrophy(TrophyType):
            category = 'old'
            image_width = 1
            image_height = 1

            def get_display_text(self, trophy):
                return 'Trophy get!'

            def qualifies(self, review_request):
                return True

        trophies_registry.register(OldTrophy)

        review_request = self.create_review_request(publish=True)

        t = Template(
            '{% load reviewtags %}'
            '{% display_review_request_trophies review_request %}')

        request = self._request_factory.get('/')
        request.user = review_request.submitter

        with self.assert_warns():
            text = t.render(RequestContext(request, {
                'review_request': review_request,
            }))

        self.assertIn('Trophy get!', text)

    def test_new_style_trophy(self):
        """Testing {% display_review_request_trophies %} for new-style
        TrophyType
        """
        class SomeTrophy(TrophyType):
            category = 'trophy'
            image_width = 1
            image_height = 1

            display_format_str = 'Trophy get!'

            def qualifies(self, review_request):
                return True

        trophies_registry.register(SomeTrophy)

        review_request = self.create_review_request(publish=True)

        t = Template(
            '{% load reviewtags %}'
            '{% display_review_request_trophies review_request %}')

        request = self._request_factory.get('/')
        request.user = review_request.submitter

        text = t.render(RequestContext(request, {
            'review_request': review_request,
        }))

        self.assertIn('Trophy get!', text)


class ForReviewRequestFieldTests(SpyAgency, TestCase):
    """Tests for the for_review_request_field template tag."""

    @add_fixtures(['test_users'])
    def test_render_instantiated_fields(self):
        """Testing for_review_request_field does not try to render
        uninstantiated fields
        """
        # exception_id will be a unique value (the ID of the field set) that
        # causes the exception; no other exception should have this value.
        exception_id = None

        class TestField(BaseReviewRequestField):
            field_id = 'test_field'

            def __init__(self, *args, **kwargs):
                raise Exception(exception_id)

        class TestFieldSet(BaseReviewRequestFieldSet):
            fieldset_id = 'test_fieldset'

        register_review_request_fieldset(TestFieldSet)
        TestFieldSet.add_field(TestField)

        review_request = self.create_review_request()

        self.spy_on(logging.exception)

        fieldset = TestFieldSet(review_request)
        exception_id = id(fieldset)

        try:
            t = Template(
                '{% load reviewtags %}'
                '{% for_review_request_field review_request fieldset %}'
                'Never reached.'
                '{% end_for_review_request_field %}'
            )

            result = t.render(Context({
                'review_request': review_request,
                'fieldset': TestFieldSet(review_request),
            }))

            self.assertEqual(result, '')
        finally:
            unregister_review_request_fieldset(TestFieldSet)

        # There should only be one logging.exception call, from the failed
        # instantiation of the TestField.
        self.assertEqual(len(logging.exception.spy.calls), 1)
        self.assertEqual(len(logging.exception.spy.calls[0].args), 3)
        self.assertEqual(
            logging.exception.spy.calls[0].args[2].args,
            (exception_id,))


class DiffCommentLineNumbersTests(TestCase):
    """Tests for the diff_comment_line_numbers template tag."""

    def test_delete_single_lines(self):
        """Testing diff_comment_line_numbers with delete chunk and single
        commented line
        """
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=1),
            'chunks': [
                {
                    'change': 'delete',
                    'lines': [
                        (10, 20, 'deleted line', [], '', '', [], False),
                        # ...
                        (50, 60, 'deleted line', [], '', '', [], False),
                    ],
                },
            ],
        }))

        self.assertEqual(result, 'Line 30 (original)')

    def test_delete_mutiple_lines(self):
        """Testing diff_comment_line_numbers with delete chunk and multiple
        commented lines
        """
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=2),
            'chunks': [
                {
                    'change': 'delete',
                    'lines': [
                        (10, 20, 'deleted line', [], '', '', [], False),
                        # ...
                        (50, 60, 'deleted line', [], '', '', [], False),
                    ],
                },
            ],
        }))

        self.assertEqual(result, 'Lines 30-31 (original)')

    def test_replace_single_line(self):
        """Testing diff_comment_line_numbers with replace chunk and single
        commented line
        """
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=1),
            'chunks': [
                {
                    'change': 'replace',
                    'lines': [
                        (10, 20, 'foo', [], 20, 'replaced line', [], False),
                        # ...
                        (50, 60, 'foo', [], 60, 'replaced line', [], False),
                    ],
                },
            ],
        }))

        self.assertEqual(result,
                         'Line 30 (original), 30 (patched)')

    def test_replace_multiple_lines(self):
        """Testing diff_comment_line_numbers with replace chunk and multiple
        commented lines
        """
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=2),
            'chunks': [
                {
                    'change': 'replace',
                    'lines': [
                        (10, 20, 'foo', [], 20, 'replaced line', [], False),
                        # ...
                        (50, 60, 'foo', [], 60, 'replaced line', [], False),
                    ],
                },
            ],
        }))

        self.assertEqual(result,
                         'Lines 30-31 (original), 30-31 (patched)')

    def test_insert_single_line(self):
        """Testing diff_comment_line_numbers with insert chunk and single
        comented line
        """
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=1),
            'chunks': [
                {
                    'change': 'insert',
                    'lines': [
                        (10, '', '', [], 20, 'inserted line', [], False),
                        # ...
                        (50, '', '', [], 60, 'inserted line', [], False),
                    ],
                },
            ],
        }))

        self.assertEqual(result, 'Lines 30 (patched)')

    def test_insert_multiple_lines(self):
        """Testing diff_comment_line_numbers with insert chunk and multiple
        commented lines
        """
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=2),
            'chunks': [
                {
                    'change': 'insert',
                    'lines': [
                        (10, '', '', [], 20, 'inserted line', [], False),
                        # ...
                        (50, '', '', [], 60, 'inserted line', [], False),
                    ],
                },
            ],
        }))

        self.assertEqual(result, 'Lines 30-31 (patched)')

    def test_fake_equal_orig(self):
        """Testing diff_comment_line_numbers with fake equal from original
        side of interdiff
        """
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=2),
            'chunks': [
                {
                    'change': 'equal',
                    'lines': [
                        (10, '', '', [], 20, 'inserted line', [], False),
                        # ...
                        (50, '', '', [], 60, 'inserted line', [], False),
                    ],
                },
            ],
        }))

        self.assertEqual(result, 'Lines 30-31 (patched)')

    def test_fake_equal_patched(self):
        """Testing diff_comment_line_numbers with fake equal from patched
        side of interdiff
        """
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=2),
            'chunks': [
                {
                    'change': 'equal',
                    'lines': [
                        (10, 20, 'deleted line', [], '', '', [], False),
                        # ...
                        (50, 60, 'deleted line', [], '', '', [], False),
                    ],
                },
            ],
        }))

        self.assertEqual(result, 'Lines 30-31 (original)')

    def test_spanning_inserts_deletes(self):
        """Testing diff_comment_line_numbers with spanning delete and insert"""
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=50),
            'chunks': [
                {
                    'change': 'delete',
                    'lines': [
                        (10, 20, 'deleted line', [], '', '', [], False),
                        # ...
                        (50, 60, 'deleted line', [], '', '', [], False),
                    ],
                },
                {
                    'change': 'insert',
                    'lines': [
                        (51, '', '', [], 61, 'inserted line', [], False),
                        # ...
                        (100, '', '', [], 110, 'inserted line', [], False),
                    ],
                },
                {
                    'change': 'equal',
                    'lines': [
                        (101, 61, 'equal line', [], 111, 'equal line', [],
                         False),
                        # ...
                        (200, 160, 'equal line', [], 210, 'equal line', [],
                         False),
                    ],
                },
            ],
        }))

        self.assertEqual(result, 'Lines 30-60 (original), 61-79 (patched)')

    def test_spanning_deletes_inserts(self):
        """Testing diff_comment_line_numbers with spanning insert and delete"""
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=50),
            'chunks': [
                {
                    'change': 'insert',
                    'lines': [
                        (10, '', '', [], 20, 'inserted line', [], False),
                        # ...
                        (50, '', '', [], 60, 'inserted line', [], False),
                    ],
                },
                {
                    'change': 'delete',
                    'lines': [
                        (51, 61, 'inserted line', [], '', '', [], False),
                        # ...
                        (100, 110, 'inserted line', [], '', '', [], False),
                    ],
                },
                {
                    'change': 'equal',
                    'lines': [
                        (101, 111, 'equal line', [], 61, 'equal line', [],
                         False),
                        # ...
                        (200, 210, 'equal line', [], 160, 'equal line', [],
                         False),
                    ],
                },
            ],
        }))

        self.assertEqual(result, 'Lines 61-79 (original), 30-60 (patched)')

    def test_spanning_last_chunk(self):
        """Testing diff_comment_line_numbers with spanning chunks through last
        chunk
        """
        t = Template(
            '{% load reviewtags %}'
            '{% diff_comment_line_numbers chunks comment %}'
        )

        result = t.render(Context({
            'comment': Comment(first_line=20, num_lines=50),
            'chunks': [
                {
                    'change': 'delete',
                    'lines': [
                        (10, 20, 'deleted line', [], '', '', [], False),
                        # ...
                        (50, 60, 'deleted line', [], '', '', [], False),
                    ],
                },
                {
                    'change': 'insert',
                    'lines': [
                        (51, '', '', [], 61, 'inserted line', [], False),
                        # ...
                        (100, '', '', [], 110, 'inserted line', [], False),
                    ],
                },
            ],
        }))

        self.assertEqual(result, 'Lines 30-60 (original), 61-79 (patched)')


class CommentRepliesTests(TestCase):
    """Unit tests for the comment_replies template tag."""

    fixtures = ['test_users']

    @add_fixtures(['test_scmtools'])
    def test_diff_comments(self):
        """Testing comment_replies for diff comments"""
        self._test_diff_comments(user_is_owner=False)

    @add_fixtures(['test_scmtools'])
    def test_diff_comments_with_draft(self):
        """Testing comment_replies for diff comments with draft"""
        self._test_diff_comments(user_is_owner=True)

    def test_general_comments(self):
        """Testing comment_replies for general comments"""
        self._test_general_comments(user_is_owner=False)

    def test_general_comments_with_draft(self):
        """Testing comment_replies for general comments with draft"""
        self._test_general_comments(user_is_owner=True)

    def test_file_attachment_comments(self):
        """Testing comment_replies for file attachment comments"""
        self._test_file_attachment_comments(user_is_owner=False)

    def test_file_attachment_comments_with_draft(self):
        """Testing comment_replies for file attachment comments with draft"""
        self._test_file_attachment_comments(user_is_owner=True)

    def test_screenshot_comments(self):
        """Testing comment_replies for screenshot comments"""
        self._test_screenshot_comments(user_is_owner=False)

    def test_screenshot_comments_with_draft(self):
        """Testing comment_replies for screenshot comments with draft"""
        self._test_screenshot_comments(user_is_owner=True)

    def _test_diff_comments(self, user_is_owner):
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request, publish=True)
        comment = self.create_diff_comment(review, filediff)

        self._check_replies(
            review,
            comment,
            self.create_diff_comment,
            {
                'filediff': filediff,
            },
            user_is_owner)

    def _test_general_comments(self, user_is_owner):
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, publish=True)
        comment = self.create_general_comment(review)

        self._check_replies(
            review,
            comment,
            self.create_general_comment,
            {},
            user_is_owner)

    def _test_file_attachment_comments(self, user_is_owner):
        review_request = self.create_review_request(publish=True)
        file_attachment = self.create_file_attachment(review_request)

        review = self.create_review(review_request, publish=True)
        comment = self.create_file_attachment_comment(review, file_attachment)

        self._check_replies(
            review,
            comment,
            self.create_file_attachment_comment,
            {
                'file_attachment': file_attachment,
            },
            user_is_owner)

    def _test_screenshot_comments(self, user_is_owner):
        review_request = self.create_review_request(publish=True)
        screenshot = self.create_screenshot(review_request)

        review = self.create_review(review_request, publish=True)
        comment = self.create_screenshot_comment(review, screenshot)

        self._check_replies(
            review,
            comment,
            self.create_screenshot_comment,
            {
                'screenshot': screenshot,
            },
            user_is_owner)

    def _check_replies(self, review, comment, create_comment_func,
                       create_comment_kwargs, user_is_owner):
        reply_kwargs = {
            'review': review,
            'user': review.user,
        }

        create_comment_kwargs['reply_to'] = comment

        reply1 = self.create_reply(publish=True, **reply_kwargs)
        reply_comment1 = create_comment_func(reply1, **create_comment_kwargs)
        reply_comment2 = create_comment_func(reply1, **create_comment_kwargs)

        reply2 = self.create_reply(publish=True, **reply_kwargs)
        reply_comment3 = create_comment_func(reply2, **create_comment_kwargs)

        reply3 = self.create_reply(publish=False, **reply_kwargs)
        reply_comment4 = create_comment_func(reply3, **create_comment_kwargs)

        t = Template(
            '{% load reviewtags %}'
            '{% comment_replies review comment "123" %}'
        )

        request = RequestFactory().request()

        if user_is_owner:
            request.user = review.user
        else:
            request.user = User.objects.create_user(username='test-user',
                                                    email='user@example.com')

        html = t.render(RequestContext(request, {
            'comment': comment,
            'review': review,
        }))

        self.assertIn('data-comment-id="%s"' % reply_comment1.pk, html)
        self.assertIn('data-comment-id="%s"' % reply_comment2.pk, html)
        self.assertIn('data-comment-id="%s"' % reply_comment3.pk, html)

        if user_is_owner:
            self.assertIn('<li class="draft" data-comment-id="%s"'
                          % reply_comment4.pk,
                          html)
        else:
            self.assertNotIn('data-comment-id="%s"' % reply_comment4.pk, html)


class ReviewBodyRepliesTests(TestCase):
    """Unit tests for the review_body_replies template tag."""

    fixtures = ['test_users']

    def test_body_top(self):
        """Testing review_body_replies for body_top"""
        self._test_body_field('body_top', user_is_owner=False)

    def test_body_top_with_draft(self):
        """Testing review_body_replies for body_top with draft"""
        self._test_body_field('body_top', user_is_owner=True)

    def test_body_bottom(self):
        """Testing review_body_replies for body_bottom"""
        self._test_body_field('body_bottom', user_is_owner=False)

    def test_body_bottom_with_draft(self):
        """Testing review_body_replies for body_bottom with draft"""
        self._test_body_field('body_bottom', user_is_owner=True)

    def _test_body_field(self, body_field, user_is_owner):
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, publish=True)

        reply_kwargs = {
            'review': review,
            'user': review.user,
            '%s_reply_to' % body_field: review,
            body_field: 'Some reply',
        }

        reply1 = self.create_reply(publish=True, **reply_kwargs)
        reply2 = self.create_reply(publish=True, **reply_kwargs)
        reply3 = self.create_reply(publish=False, **reply_kwargs)

        t = Template(
            '{%% load reviewtags %%}'
            '{%% review_body_replies review "%s" "123" %%}'
            % body_field
        )

        request = RequestFactory().request()

        if user_is_owner:
            request.user = review.user
        else:
            request.user = User.objects.create_user(username='test-user',
                                                    email='user@example.com')

        html = t.render(RequestContext(request, {
            'review': review,
        }))

        self.assertIn('id="comment_123-%s"' % reply1.pk, html)
        self.assertIn('id="comment_123-%s"' % reply2.pk, html)

        if user_is_owner:
            self.assertIn('id="draftcomment_123-%s"' % reply3.pk, html)
        else:
            self.assertNotIn('id="comment_123-%s"' % reply3.pk, html)
            self.assertNotIn('id="draftcomment_123-%s"' % reply3.pk, html)



"""Views for serving up attachments."""

from __future__ import unicode_literals

from django.shortcuts import get_object_or_404, redirect
from django.views.static import serve

from reviewboard.attachments.models import FileAttachment
from reviewboard.site.decorators import check_local_site_access


@check_local_site_access
def user_file_attachment(request, file_attachment_uuid, username,
                         local_site=None):
    """Redirect to the file attachment's URL given its UUID.

    Args:
        request (django.http.HttpRequest):
            The request.

        file_attachment_uuid (unicode):
            The UUID of the file attachment.

        username (unicode):
            The username of the user who uploaded the file.

        local_site (reviewboard.site.models.LocalSite, optional):
            The local site, if any.

    Returns:
        django.http.HttpResponseRedirect:
        The response to send back to the browser.
    """
    file_attachment = get_object_or_404(FileAttachment,
                                        uuid=file_attachment_uuid,
                                        user__username=username,
                                        local_site=local_site,
                                        file__isnull=False)

    return redirect(file_attachment)


def serve_safe(*args, **kwargs):
    """Safely serve an uploaded file to the client.

    This is a wrapper around :py:func:`django.views.static.serve`, intended
    only for use in DEBUG mode, that will serve files to the client using the
    security rules recommended for production use. All files will be returned
    with a ``Content-Disposition: attachment`` header, ensuring that the
    browser will download the file instead of attempting to open it inline.

    Args:
        *args (tuple):
            Additional positional arguments.

        **kwargs (dict):
            Additional keyword arguments.

    Returns:
        django.http.HttpResponse:
        The HTTP response for the file.
    """
    response = serve(*args, **kwargs)

    if response.status_code == 200:
        response['Content-Disposition'] = 'attachment'

    return response



"""Feature definitions for reviews."""

from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _
from djblets.features import Feature, FeatureLevel


class ClassBasedActionsFeature(Feature):
    """A feature for class-based actions.

    With this enabled, extensions may use the new class-based action classes
    instead of the old-style dict actions.
    """

    feature_id = 'reviews.class_based_actions'
    name = _('Class-Based Actions')
    level = FeatureLevel.UNAVAILABLE
    summary = _('Allow using class-based actions with extension hooks.')


class GeneralCommentsFeature(Feature):
    """A feature for general comments.

    General comments allow comments to be created directly on a review request
    without accompanying file attachment or diff. These can be used to raise
    issues with the review request itself, such as its summary or description,
    or general implementation issues.
    """

    feature_id = 'reviews.general_comments'
    name = _('General Comments')
    level = FeatureLevel.STABLE
    summary = _('Allow comments on review requests without an associated file '
                'attachment or diff.')


class IssueVerificationFeature(Feature):
    """A feature for issue verification.

    Issue verification allows reviewers to mark that an issue requires
    verification before closing. In this case, the author of the change will be
    able to mark the issue as "Fixed", but then the original author of the
    comment will need to verify it before the issue is closed.
    """

    feature_id = 'reviews.issue_verification'
    name = _('Issue Verification')
    level = FeatureLevel.STABLE
    summary = _('Allow comment authors to require that issues be verified by '
                'them before being closed')


class StatusUpdatesFeature(Feature):
    """A feature for status updates.

    A status update is a way for third-party tools to provide feedback on a
    review request. In the past, this was done just as a normal review. Status
    updates allow those tools (via some integration like Review Bot) to mark
    their state (such as pending, success, failure, or error) and then
    associate that with a review.
    """

    feature_id = 'reviews.status_updates'
    name = _('Status Updates')
    level = FeatureLevel.STABLE
    summary = _('A way for external tools to do checks on a review request '
                'and report the results of those checks.')


class_based_actions_feature = ClassBasedActionsFeature()
general_comments_feature = GeneralCommentsFeature()
issue_verification_feature = IssueVerificationFeature()
status_updates_feature = StatusUpdatesFeature()



from __future__ import unicode_literals

import warnings

from reviewboard.deprecation import RemovedInReviewBoard40Warning
from reviewboard.hostingsvcs.testing import HostingServiceTestCase


class ServiceTests(HostingServiceTestCase):
    """Legacy base class for hosting service tests.

    .. deprecated:: 3.0.4

       This is an old class and should no longer be used as the base class
       for hosting service test suites. Use :py:class:`HostingServiceTestCase`
       instead.
    """

    @classmethod
    def setUpClass(cls):
        warnings.warn('ServiceTests is deprecated. Subclass '
                      'HostingServiceTestCase instead.',
                      RemovedInReviewBoard40Warning)

        super(ServiceTests, cls).setUpClass()

    # Legacy private functions previously used by subclasses.
    #
    # These have never been documented and still aren't. They shouldn't be
    # used, technically, and will go away in time.
    def _get_form(self, *args, **kwargs):
        warnings.warn(
            'ServiceTests._get_form() is deprecated. Use '
            'HostingServiceTestCase.get_form() instead.',
            RemovedInReviewBoard40Warning)

        return self.get_form(*args, **kwargs)

    def _get_hosting_account(self, *args, **kwargs):
        warnings.warn(
            'ServiceTests._get_hosting_account() is deprecated. Use '
            'HostingServiceTestCase.create_hosting_account() instead.',
            RemovedInReviewBoard40Warning)

        kwargs['data'] = {}
        return self.create_hosting_account(*args, **kwargs)

    def _get_repository_fields(self, *args, **kwargs):
        warnings.warn(
            'ServiceTests._get_repository_fields() is deprecated. Use '
            'HostingServiceTestCase.get_repository_fields() instead.',
            RemovedInReviewBoard40Warning)

        return self.get_repository_fields(*args, **kwargs)



"""The Review Board e-mail message class and methods for generating e-mails."""

from __future__ import unicode_literals

import logging

from django.conf import settings
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.template.loader import render_to_string
from django.utils import six
from django.utils.datastructures import MultiValueDict
from djblets.mail.message import EmailMessage as DjbletsEmailMessage
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.accounts.pages import AuthenticationPage
from reviewboard.admin.server import build_server_url, get_server_url
from reviewboard.notifications.email.hooks import \
    filter_email_recipients_from_hooks
from reviewboard.notifications.email.utils import (
    build_email_address,
    build_email_address_for_user,
    build_recipients,
    recipients_to_addresses)
from reviewboard.reviews.models import Group
from reviewboard.reviews.signals import (review_request_published,
                                         review_published, reply_published,
                                         review_request_closed)


MAX_FILENAME_HEADERS_LENGTH = 8192

#: The number of additional characters each ``X-ReviewBoard-Diff-For`` has.
#:
#: We calculate the length the value of each header at runtime. However,
#: ``X-ReviewBoard-Diff-For: `` is present before the value, and the line
#: terminates with a ``\r\n``.
HEADER_ADDITIONAL_CHARACTERS_LENGTH = (len(b'\r\n') +
                                       len(b'X-ReviewBoard-Diff-For: '))


class EmailMessage(DjbletsEmailMessage):
    """An e-mail message.

    This class only differs from Djblets'
    :py:class:`~djblets.email.message.EmailMessage` by using the site
    configuration to generate some e-mail settings.
    """

    def __init__(self, *args, **kwargs):
        siteconfig = SiteConfiguration.objects.get_current()
        auto_generated = siteconfig.get('mail_enable_autogenerated_header')

        super(EmailMessage, self).__init__(
            auto_generated=auto_generated,
            prevent_auto_responses=True,
            *args,
            **kwargs)


def _ensure_unicode(text):
    """Return a unicode object for the given text.

    Args:
        text (bytes or unicode):
            The text to decode.

    Returns:
        unicode: The decoded text.
    """
    if isinstance(text, bytes):
        text = text.decode('utf-8')

    return text


def _get_server_base_url():
    """Return the base URL of the server (without a trailing /).

    Returns:
        unicode:
        The server base URL without a trailing slash.

        For a site at :samp:`{scheme}://example.com/site-root`, this function
        will return :samp:`{scheme}://example.com`.
    """
    return build_server_url('/')[:-1]


def prepare_base_review_request_mail(user, review_request, subject,
                                     in_reply_to, to_field, cc_field,
                                     template_name_base, context=None,
                                     extra_headers=None):
    """Return a customized review request e-mail.

    This is intended to be called by one of the ``prepare_{type}_mail``
    functions in this file. This method builds up a common context that all
    review request-related e-mails will use to render their templates, as well
    as handling user preferences regarding e-mail and add adding additional
    headers.

    Args:
        user (django.contrib.auth.models.User):
            The user who is sending the e-mail.

        review_request (reviewboard.reviews.models.review_request.ReviewRequest):
            The review request this e-mail is regarding.

        subject (unicode):
            The e-mail subject line.

        in_reply_to (unicode):
            The e-mail message ID this message is in response to or ``None``.

        to_field (set):
            The set of :py:class:`~django.contrib.auth.models.User` and
            :py:class`~reviewboard.reviews.models.group.Group`s to this e-mail
            will be sent to.

        cc_field (set):
            The set of :py:class:`~django.contrib.auth.models.User` and
            :py:class`~reviewboard.reviews.models.group.Group`s to be CC'ed on
            the e-mail.

        template_name_base (unicode):
            The name of the template to use to generate the e-mail without its
            extension. The plain-text version of the e-mail will append
            ``.txt`` to this and and the rich-text version of the e-mail will
            append ``.html``.

        context (dict, optional):
            Optional additional template rendering context.

        extra_headers (dict, optional):
            Optional additional headers to include.

    Returns:
        EmailMessage:
        The prepared e-mail message.
    """
    user_email = build_email_address_for_user(user)
    to_field = recipients_to_addresses(to_field, review_request.id)
    cc_field = recipients_to_addresses(cc_field, review_request.id) - to_field

    if not user.should_send_own_updates():
        to_field.discard(user_email)
        cc_field.discard(user_email)

    if not to_field and not cc_field:
        # This e-mail would have no recipients, so we won't send it.
        return None

    if not context:
        context = {}

    context.update({
        'user': user,
        'site_url': _get_server_base_url(),
        'review_request': review_request,
    })
    local_site = review_request.local_site

    if local_site:
        context['local_site_name'] = local_site.name

    text_body = render_to_string('%s.txt' % template_name_base, context)
    html_body = render_to_string('%s.html' % template_name_base, context)
    server_url = get_server_url(local_site=local_site)

    headers = MultiValueDict({
        'X-ReviewBoard-URL': [server_url],
        'X-ReviewRequest-URL': [
            build_server_url(review_request.get_absolute_url(),
                             local_site=local_site)
        ],
        'X-ReviewGroup': [', '.join(
            review_request.target_groups.values_list('name', flat=True)
        )],
    })

    if extra_headers:
        if not isinstance(extra_headers, MultiValueDict):
            extra_headers = MultiValueDict(
                (key, [value])
                for key, value in six.iteritems(extra_headers)
            )

        headers.update(extra_headers)

    if review_request.repository:
        headers['X-ReviewRequest-Repository'] = review_request.repository.name

    latest_diffset = review_request.get_latest_diffset()

    if latest_diffset:
        modified_files = set()

        for filediff in latest_diffset.files.all():
            if not filediff.is_new:
                modified_files.add(filediff.source_file)

            if not filediff.deleted:
                modified_files.add(filediff.dest_file)

        # The following code segment deals with the case where the client adds
        # a significant amount of files with large names. We limit the number
        # of headers; when more than 8192 characters are reached, we stop
        # adding filename headers.
        current_header_length = 0

        for filename in modified_files:
            current_header_length += (HEADER_ADDITIONAL_CHARACTERS_LENGTH +
                                      len(filename))

            if current_header_length > MAX_FILENAME_HEADERS_LENGTH:
                logging.warning(
                    'Unable to store all filenames in the '
                    'X-ReviewBoard-Diff-For headers when sending e-mail for '
                    'review request %s: The header size exceeds the limit of '
                    '%s. Remaining headers have been omitted.',
                    review_request.display_id,
                    MAX_FILENAME_HEADERS_LENGTH)
                break

            headers.appendlist('X-ReviewBoard-Diff-For', filename)

    if settings.DEFAULT_FROM_EMAIL:
        sender = build_email_address(full_name=user.get_full_name(),
                                     email=settings.DEFAULT_FROM_EMAIL)
    else:
        sender = None

    return EmailMessage(subject=subject.strip(),
                        text_body=text_body.encode('utf-8'),
                        html_body=html_body.encode('utf-8'),
                        from_email=user_email,
                        sender=sender,
                        to=list(to_field),
                        cc=list(cc_field),
                        in_reply_to=in_reply_to,
                        headers=headers)


def prepare_password_changed_mail(user):
    """Return an e-mail notifying the user that their password changed.

    Args:
        user (django.contrib.auth.models.User):
            The user whose password changed.

    Returns:
        EmailMessage:
        The generated message.
    """
    server_url = get_server_url()

    context = {
        'api_token_url': AuthenticationPage.get_absolute_url(),
        'has_api_tokens': user.webapi_tokens.exists(),
        'server_url': server_url,
        'user': user,
    }

    user_email = build_email_address_for_user(user)
    text_body = render_to_string('notifications/password_changed.txt', context)
    html_body = render_to_string('notifications/password_changed.html',
                                 context)

    return EmailMessage(
        subject='Password changed for user "%s" on %s' % (user.username,
                                                          server_url),
        text_body=text_body,
        html_body=html_body,
        from_email=settings.SERVER_EMAIL,
        sender=settings.SERVER_EMAIL,
        to=user_email)


def prepare_reply_published_mail(user, reply, review, review_request):
    """Return an e-mail representing the supplied reply to a review.

    Args:
        user (django.contrib.auth.models.User):
            The user who published the reply.

        reply (reviewboard.reviews.models.review.Review):
            The review reply to send an e-mail about.

        review (reviewboard.reviews.models.review.Review):
            The review that was replied to.

        review_request (reviewboard.reviews.models.review_request.ReviewRequest):
            The review request.

    Returns:
        EmailMessage:
        The generated e-mail message.
    """
    from reviewboard.reviews.views import build_diff_comment_fragments

    extra_context = {
        'user': reply.user,
        'review': review,
        'reply': reply,
        'site_url': _get_server_base_url(),
    }

    extra_context['comment_entries'] = build_diff_comment_fragments(
        reply.comments.order_by('filediff', 'first_line'),
        extra_context,
        'notifications/email_diff_comment_fragment.html')[1]

    to_field, cc_field = build_recipients(reply.user, review_request,
                                          review_request.participants)

    to_field, cc_field = filter_email_recipients_from_hooks(
        to_field, cc_field, reply_published,
        reply=reply,
        user=user,
        review=review,
        review_request=review_request)

    summary = _ensure_unicode(review_request.summary)

    return prepare_base_review_request_mail(
        user, review_request,
        'Re: Review Request %d: %s' % (review_request.display_id, summary),
        review.email_message_id, to_field, cc_field,
        'notifications/reply_email', extra_context)


def prepare_review_published_mail(user, review, review_request, request,
                                  to_owner_only=False):
    """Return an e-mail representing the supplied review.

    Args:
        user (django.contrib.auth.models.User):
            The user who published the review.

        review (reviewboard.reviews.models.review.Review):
            The review to send an e-mail about.

        review_request (reviewboard.reviews.models.review_request.ReviewRequest):
            The review request that was reviewed.

        to_owner_only (bool):
            Whether or not the review should be sent to the submitter only.

    Returns:
        EmailMessage:
        The generated e-mail message.
    """
    from reviewboard.reviews.views import build_diff_comment_fragments

    review.ordered_comments = review.comments.order_by('filediff',
                                                       'first_line')
    has_issues = (review.ship_it and
                  review.has_comments(only_issues=True))
    extra_context = {
        'user': review.user,
        'review': review,
        'has_issues': has_issues,
        'request': request,
        'site_url': _get_server_base_url(),
    }

    extra_headers = {}

    if review.ship_it:
        extra_headers['X-ReviewBoard-ShipIt'] = '1'

        if review.ship_it_only:
            extra_headers['X-ReviewBoard-ShipIt-Only'] = '1'

    extra_context['comment_entries'] = build_diff_comment_fragments(
        review.ordered_comments, extra_context,
        'notifications/email_diff_comment_fragment.html')[1]

    limit_to = None

    if to_owner_only:
        limit_to = {review_request.submitter, review.user}

    to_field, cc_field = build_recipients(review.user, review_request,
                                          limit_recipients_to=limit_to)

    to_field, cc_field = filter_email_recipients_from_hooks(
        to_field, cc_field, review_published,
        user=user,
        review=review,
        to_owner_only=to_owner_only,
        review_request=review_request)

    summary = _ensure_unicode(review_request.summary)

    return prepare_base_review_request_mail(
        review.user, review_request,
        'Re: Review Request %d: %s' % (review_request.display_id, summary),
        review_request.email_message_id, to_field, cc_field,
        'notifications/review_email', extra_context,
        extra_headers=extra_headers)


def prepare_review_request_mail(user, review_request, changedesc=None,
                                close_type=None):
    """Return an e-mail representing the supplied review request.

    Args:
        user (django.contrib.auth.models.User):
            The user who triggered the e-mail (i.e., they published or
            closed he review request).

        review_request (reviewboard.reviews.models.ReviewRequest):
            The review request to send an e-mail about.

        changedesc (reviewboard.changedescs.models.ChangeDescription):
            An optional change description showing what has changed in the
            review request, possibly with explanatory text from the
            submitter. This is created when saving a draft on a public review
            request and will be ``None`` when publishing initially. This is
            used by the template to add contextual (updated) flags to inform
            people what has changed.

        close_type (unicode):
            How the review request was closed or ``None`` if it was
            published. If this is not ``None`` it must be one of:

            * :py:attr:`~reviewboard.reviews.models.ReviewRequest.SUBMITTED`
            * :py:attr:`~reviewboard.reviews.models.ReviewRequest.DISCARDED`

    Returns:
        EmailMessage:
        The e-mail message representing the review request.
    """
    if not user:
        user = review_request.submitter

    summary = _ensure_unicode(review_request.summary)
    subject = 'Review Request %d: %s' % (review_request.display_id, summary)

    reply_message_id = review_request.email_message_id
    extra_recipients = None

    if reply_message_id:
        # Fancy quoted "replies".
        subject = 'Re: %s' % subject
        extra_recipients = review_request.participants

    extra_context = {}
    extra_filter_kwargs = {}

    if close_type:
        changedesc = review_request.changedescs.filter(public=True).latest()
        signal = review_request_closed
        extra_filter_kwargs['close_type'] = close_type
    else:
        signal = review_request_published

    limit_recipients_to = None

    if changedesc:
        fields_changed = changedesc.fields_changed
        changed_field_names = set(fields_changed)
        extra_context.update({
            'change_text': changedesc.text,
            'change_rich_text': changedesc.rich_text,
            'changes': fields_changed,
        })

        if (changed_field_names and
            changed_field_names <= {'target_people', 'target_groups'}):
            # If the only changes are to the target reviewers, try to send a
            # much more targeted e-mail. Rather than having it be sent out to
            # everyone, it will only be sent to new reviewers.
            limit_recipients_to = set()

            for model, field in ((User, 'target_people'),
                                 (Group, 'target_groups')):
                if field in changed_field_names:
                    limit_recipients_to.update(
                        model.objects.filter(pk__in=[
                            item[2]
                            for item in fields_changed[field]['added']
                        ]))

    to_field, cc_field = build_recipients(
        user, review_request, extra_recipients,
        limit_recipients_to=limit_recipients_to)

    to_field, cc_field = filter_email_recipients_from_hooks(
        to_field, cc_field, signal,
        review_request=review_request,
        user=user,
        **extra_filter_kwargs)

    return prepare_base_review_request_mail(
        user, review_request, subject, reply_message_id, to_field,
        cc_field, 'notifications/review_request_email', extra_context)


def prepare_user_registered_mail(user):
    """Prepare an e-mail to the administrators notifying of a new user.

    Args:
        user (django.contrib.auth.models.User):
            The user who registered.

    Returns:
        EmailMessage:
        The generated e-mail.
    """
    subject = "New Review Board user registration for %s" % user.username

    context = {
        'site_url': _get_server_base_url(),
        'user': user,
        'user_url': build_server_url(reverse('admin:auth_user_change',
                                             args=(user.id,))),
    }

    text_message = render_to_string('notifications/new_user_email.txt',
                                    context)
    html_message = render_to_string('notifications/new_user_email.html',
                                    context)

    return EmailMessage(
        subject=subject.strip(),
        text_body=text_message,
        html_body=html_message,
        from_email=settings.SERVER_EMAIL,
        sender=settings.SERVER_EMAIL,
        to=[
            build_email_address(full_name=admin[0], email=admin[1])
            for admin in settings.ADMINS
        ])


def prepare_webapi_token_mail(webapi_token, op):
    """Return an e-mail message notifying a user about a WebAPI token change.

    Args:
        webapi_token (reviewboard.notifications.models.WebAPIToken):
            The token that was created, updated, or deleted.

        op (unicode):
            The operation on the token. This is one of:

            * ``'created'``
            * ``'updated'``
            * ``'deleted'``

    Returns:
        EmailMessage:
        The genereated e-mail.
    """
    if op == 'created':
        subject = 'New Review Board API token created'
        template_name = 'notifications/api_token_created'
    elif op == 'updated':
        subject = 'Review Board API token updated'
        template_name = 'notifications/api_token_updated'
    elif op == 'deleted':
        subject = 'Review Board API token deleted'
        template_name = 'notifications/api_token_deleted'
    else:
        raise ValueError('Unexpected op "%s" passed to mail_webapi_token.'
                         % op)

    user = webapi_token.user
    user_email = build_email_address_for_user(user)

    context = {
        'api_token': webapi_token,
        'api_tokens_url': AuthenticationPage.get_absolute_url(),
        'partial_token': '%s...' % webapi_token.token[:10],
        'user': user,
        'site_root_url': get_server_url(),
    }

    text_message = render_to_string('%s.txt' % template_name, context)
    html_message = render_to_string('%s.html' % template_name, context)

    return EmailMessage(
        subject=subject,
        text_body=text_message,
        html_body=html_message,
        from_email=settings.SERVER_EMAIL,
        sender=settings.SERVER_EMAIL,
        to=[user_email])



from __future__ import unicode_literals


class OwnershipError(ValueError):
    """An error that occurs when a user does not own a review request."""
    pass


class PermissionError(Exception):
    """An error that occurs when a user does not have required permissions."""
    pass


class PublishError(Exception):
    """An error that occurs when attempting to publish.

    The model triggering this error may be a review request, review, or reply.
    """

    def __init__(self, message):
        super(PublishError, self).__init__('Error publishing: %s' % message)


class CloseError(Exception):
    """An error that occurs while attempting to close a review request."""

    def __init__(self, message):
        super(CloseError, self).__init__(
            'Error closing the review request: %s' % message)


class ReopenError(Exception):
    """An error that occurs while attempting to reopen a review request."""

    def __init__(self, message):
        super(ReopenError, self).__init__(
            'Error reopening the review request: %s' % message)


class RevokeShipItError(Exception):
    """An error that occurs while attempting to revoke a Ship It."""

    def __init__(self, message):
        super(RevokeShipItError, self).__init__(
            'Error revoking the Ship It: %s' % message)


class NotModifiedError(PublishError):
    """An error that occurs when a review's state is not modified."""

    def __init__(self):
        super(NotModifiedError, self).__init__(
            'The draft has no modifications.')


class DepthLimitExceededError(ValueError):
    """An error that occurs when the maximum depth limit is exceeded.

    Review request actions cannot be arbitrarily nested. For example, if the
    depth limit is 2, then this error would be triggered if an extension tried
    to add a menu action as follows:

    .. code-block:: python

       BaseReviewRequestActionHook(self, actions=[
           DepthZeroMenuAction([
               DepthOneFirstItemAction(),
               DepthOneMenuAction([
                   DepthTwoMenuAction([  # This depth is acceptable.
                       DepthThreeTooDeepAction(),  # This action is too deep.
                   ]),
               ]),
               DepthOneLastItemAction(),
           ]),
       ])
    """

    def __init__(self, action_id, depth_limit):
        super(DepthLimitExceededError, self).__init__(
            '%s exceeds the maximum depth limit of %d'
            % (action_id, depth_limit))



# rbssh plugin for Bazaar.
# Copyright (C) 2017  Beanbag, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""rbssh plugin for Bazaar.

This registers :command:`rbssh` as a SSH plugin for Bazaar. It's run entirely
outside of the Review Board process, and is invoked exclusively by
:command:`bzr`.
"""

from __future__ import unicode_literals

from bzrlib.transport.ssh import (SubprocessVendor,
                                  register_default_ssh_vendor,
                                  register_ssh_vendor)
from django.utils import six


class RBSSHVendor(SubprocessVendor):
    """SSH vendor class for using rbssh."""

    executable_path = 'rbssh'

    def _get_vendor_specific_argv(self, username, host, port=None,
                                  subsystem=None, command=None):
        """Return arguments to pass to rbssh.

        Args:
            username (unicode):
                The username to connect with.

            host (unicode):
                The hostname to connect to.

            port (int, optional):
                The custom port to connect to.

            subsystem (unicode, optional):
                The SSH subsystem to use.

            command (unicode, optional):
                The command to invoke through the SSH connection.

        Returns:
            list of unicode:
            The list of arguments to pass to :command:`rbssh`.
        """
        args = [self.executable_path]

        if port is not None:
            args.extend(['-p', six.text_type(port)])

        if username is not None:
            args.extend(['-l', username])

        if subsystem is not None:
            args.extend(['-s', host, subsystem])
        else:
            args.extend([host] + command)

        return args


vendor = RBSSHVendor()
register_ssh_vendor('rbssh', vendor)
register_default_ssh_vendor(vendor)



from __future__ import unicode_literals

from djblets.registries.mixins import ExceptionFreeGetterMixin
from djblets.registries.registry import (
    EntryPointRegistry as DjbletsEntryPointRegistry,
    OrderedRegistry as DjbletsOrderedRegistry,
    Registry as DjbletsRegistry)


class Registry(ExceptionFreeGetterMixin, DjbletsRegistry):
    """A registry that does not throw exceptions for failed lookups."""


class EntryPointRegistry(ExceptionFreeGetterMixin, DjbletsEntryPointRegistry):
    """A registry that auto-populates from an entry-point."""


class OrderedRegistry(ExceptionFreeGetterMixin, DjbletsOrderedRegistry):
    """A registry that keeps track of registration order."""



from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser, User
from django.core.files.storage import get_storage_class
from django.core.files.uploadedfile import SimpleUploadedFile
from django.template import RequestContext, Template
from django.test.client import RequestFactory
from django.utils.html import escape
from djblets.avatars.services import (GravatarService,
                                      URLAvatarService)
from djblets.avatars.tests import DummyAvatarService, DummyHighDPIAvatarService
from djblets.siteconfig.models import SiteConfiguration
from kgb import SpyAgency

from reviewboard.accounts.forms.pages import AvatarSettingsForm
from reviewboard.accounts.models import Profile
from reviewboard.avatars import avatar_services
from reviewboard.avatars.registry import AvatarServiceRegistry
from reviewboard.avatars.services import FileUploadService
from reviewboard.avatars.testcase import AvatarServicesTestMixin
from reviewboard.testing.testcase import TestCase


# This just helps keep things less wordy, and to define some settings
# we no longer have constants for elsewhere.
AVATARS_ENABLED_KEY = AvatarServiceRegistry.AVATARS_ENABLED_KEY
DEFAULT_SERVICE_KEY = AvatarServiceRegistry.DEFAULT_SERVICE_KEY
ENABLED_SERVICES_KEY = AvatarServiceRegistry.ENABLED_SERVICES_KEY
LEGACY_AVATARS_MIGRATED_KEY = AvatarServiceRegistry.LEGACY_AVATARS_MIGRATED_KEY
LEGACY_INTEGRATION_GRAVATARS_KEY = \
    AvatarServiceRegistry.LEGACY_INTEGRATION_GRAVATARS_KEY


class AvatarServiceRegistryTests(AvatarServicesTestMixin, TestCase):
    """Tests for reviewboard.avatars."""

    _enabled_service_ids = [
        GravatarService.avatar_service_id,
        FileUploadService.avatar_service_id,
        URLAvatarService.avatar_service_id,
    ]

    _default_service_id = GravatarService.avatar_service_id

    def setUp(self):
        super(AvatarServiceRegistryTests, self).setUp()

        self.registry = AvatarServiceRegistry()
        self.siteconfig = SiteConfiguration.objects.get_current()

    def test_defaults(self):
        """Testing AvatarServiceRegistry default services state"""
        self.assertEqual(
            set(self.registry),
            {
                FileUploadService,
                GravatarService,
                URLAvatarService,
            })

        self.assertIsInstance(self.registry.default_service, GravatarService)
        self.assertEqual(
            set(self.registry.enabled_services),
            {
                FileUploadService,
                GravatarService,
                URLAvatarService,
            })

    def test_get_siteconfig_defaults(self):
        """Testing AvatarServiceRegistry.get_siteconfig_defaults"""
        self.assertEqual(
            self.registry.get_siteconfig_defaults(),
            {
                AVATARS_ENABLED_KEY: True,
                ENABLED_SERVICES_KEY: self._enabled_service_ids,
                DEFAULT_SERVICE_KEY: self._default_service_id,
            })

    def test_migrate_settings_pre_2_0_with_enabled(self):
        """Testing AvatarServiceRegistry.migrate_settings migrates pre-2.0
        settings
        """
        siteconfig_settings = self.siteconfig.settings

        self.assertNotIn(AVATARS_ENABLED_KEY, siteconfig_settings)
        self.assertNotIn(DEFAULT_SERVICE_KEY, siteconfig_settings)
        self.assertNotIn(ENABLED_SERVICES_KEY, siteconfig_settings)
        self.assertNotIn(LEGACY_AVATARS_MIGRATED_KEY, siteconfig_settings)
        self.assertNotIn(LEGACY_INTEGRATION_GRAVATARS_KEY, siteconfig_settings)

        # Perform the migration.
        self._perform_migration(expect_has_enabled_saved=False,
                                expect_migrated=False)

    def test_migrate_settings_pre_3_0_with_enabled(self):
        """Testing AvatarServiceRegistry.migrate_settings migrates pre-3.0
        avatar settings for enabled Gravatars
        """
        siteconfig_settings = self.siteconfig.settings

        self.assertNotIn(AVATARS_ENABLED_KEY, siteconfig_settings)
        self.assertNotIn(DEFAULT_SERVICE_KEY, siteconfig_settings)
        self.assertNotIn(ENABLED_SERVICES_KEY, siteconfig_settings)
        self.assertNotIn(LEGACY_AVATARS_MIGRATED_KEY, siteconfig_settings)

        self.siteconfig.set(LEGACY_INTEGRATION_GRAVATARS_KEY, True)

        # Perform the migration.
        self._perform_migration(expect_has_enabled_saved=False)

    def test_migrate_pre_3_0_disabled(self):
        """Testing AvatarServiceRegistry migrates pre-3.0 avatar settings for
        disabled Gravatars
        """
        siteconfig_settings = self.siteconfig.settings

        self.assertNotIn(AVATARS_ENABLED_KEY, siteconfig_settings)
        self.assertNotIn(DEFAULT_SERVICE_KEY, siteconfig_settings)
        self.assertNotIn(ENABLED_SERVICES_KEY, siteconfig_settings)
        self.assertNotIn(LEGACY_AVATARS_MIGRATED_KEY, siteconfig_settings)

        self.siteconfig.set(LEGACY_INTEGRATION_GRAVATARS_KEY, False)

        # Perform the migration.
        self._perform_migration(expect_enabled=False)

    def test_migrate_3_0_pre_4_disabled(self):
        """Testing AvatarServiceRegistry migrates 3.0.[0-3] avatar settings for
        migrated disabled avatars
        """
        enabled_service_ids = [
            GravatarService.avatar_service_id,
            URLAvatarService.avatar_service_id,
        ]
        default_service_id = URLAvatarService.avatar_service_id

        self.siteconfig.set(AVATARS_ENABLED_KEY, False)
        self.siteconfig.set(DEFAULT_SERVICE_KEY, default_service_id)
        self.siteconfig.set(ENABLED_SERVICES_KEY, enabled_service_ids)
        self.siteconfig.set(LEGACY_AVATARS_MIGRATED_KEY, True)
        self.siteconfig.set(LEGACY_INTEGRATION_GRAVATARS_KEY, True)

        # Perform the migration.
        self._perform_migration(expect_enabled=False,
                                expected_enabled_services=enabled_service_ids,
                                expected_default_service=default_service_id,
                                expect_has_services_saved=True)

    def test_migrate_3_0_pre_4_enabled(self):
        """Testing AvatarServiceRegistry migrates 3.0.[0-3] avatar settings for
        migrated enabled avatars
        """
        enabled_service_ids = [
            GravatarService.avatar_service_id,
            URLAvatarService.avatar_service_id,
        ]
        default_service_id = URLAvatarService.avatar_service_id

        self.siteconfig.set(AVATARS_ENABLED_KEY, True)
        self.siteconfig.set(DEFAULT_SERVICE_KEY, default_service_id)
        self.siteconfig.set(ENABLED_SERVICES_KEY, enabled_service_ids)
        self.siteconfig.set(LEGACY_AVATARS_MIGRATED_KEY, True)
        self.siteconfig.set(LEGACY_INTEGRATION_GRAVATARS_KEY, True)

        # Perform the migration.
        self._perform_migration(expected_enabled_services=enabled_service_ids,
                                expected_default_service=default_service_id,
                                expect_has_services_saved=True)

    def test_migrate_already_migrated_enabled(self):
        """Testing AvatarServiceRegistry does not migrate already-migrated
        settings with enabled avatars
        """
        siteconfig_settings = self.siteconfig.settings

        self.assertNotIn(DEFAULT_SERVICE_KEY, siteconfig_settings)
        self.assertNotIn(ENABLED_SERVICES_KEY, siteconfig_settings)

        self.siteconfig.set(AVATARS_ENABLED_KEY, True)
        self.siteconfig.set(LEGACY_AVATARS_MIGRATED_KEY, True)
        self.siteconfig.set(LEGACY_INTEGRATION_GRAVATARS_KEY, True)

        # Perform the migration.
        self._perform_migration()

    def test_migrate_already_migrated_disabled(self):
        """Testing AvatarServiceRegistry does not migrate already-migrated
        settings with disabled avatars
        """
        siteconfig_settings = self.siteconfig.settings

        self.assertNotIn(DEFAULT_SERVICE_KEY, siteconfig_settings)
        self.assertNotIn(ENABLED_SERVICES_KEY, siteconfig_settings)

        self.siteconfig.set(AVATARS_ENABLED_KEY, True)
        self.siteconfig.set(LEGACY_AVATARS_MIGRATED_KEY, True)
        self.siteconfig.set(LEGACY_INTEGRATION_GRAVATARS_KEY, True)

        # Perform the migration.
        self._perform_migration()

    def test_migrate_with_fresh_install(self):
        """Testing AvatarServiceRegistry does not migrate fresh settings
        settings with disabled avatars
        """
        siteconfig_settings = self.siteconfig.settings

        self.assertNotIn(AVATARS_ENABLED_KEY, siteconfig_settings)
        self.assertNotIn(DEFAULT_SERVICE_KEY, siteconfig_settings)
        self.assertNotIn(ENABLED_SERVICES_KEY, siteconfig_settings)
        self.assertNotIn(LEGACY_AVATARS_MIGRATED_KEY, siteconfig_settings)
        self.assertNotIn(LEGACY_INTEGRATION_GRAVATARS_KEY, siteconfig_settings)

        # Perform the migration.
        self._perform_migration(expect_migrated=False,
                                expect_has_enabled_saved=False)

    def _perform_migration(self,
                           expect_migrated=True,
                           expect_enabled=True,
                           expected_enabled_services=_enabled_service_ids,
                           expected_default_service=_default_service_id,
                           expect_has_services_saved=False,
                           expect_has_enabled_saved=True):
        """Perform an avatar settings migration test.

        This will trigger a migration of the existing settings, checking the
        final results in the database based on the expectations of the caller.
        This helps to unify all of our various migration tests.

        Args:
            expect_migrated (bool, optional):
                Whether the migration attempt is expected to have made changes
                to the stored site configuration.

            expect_enabled (bool, optional):
                Whether avatars are expected to be enabled after the migration.

            expected_enabled_services (list, optional):
                The list of avatar service IDs that are expected to be enabled
                after the migration.

            expected_default_service (unicode, optional):
                The avatar service ID that is expected to be set as the
                default.

            expect_has_services_saved (bool, optional):
                Whether the site configuration is expected to have a list of
                service IDs explicitly saved.

            expect_has_enabled_saved (bool, optional):
                Whether the site configuration is expected to have the enabled
                state explicitly saved.
        """
        migrated = self.registry.migrate_settings(self.siteconfig)
        self.assertEqual(migrated, expect_migrated)

        if migrated:
            self.siteconfig.save()

        # Make sure the default avatar and available/enabled avatars are
        # still there.
        self.assertEqual(self.registry.default_service.avatar_service_id,
                         expected_default_service)
        self.assertEqual(
            set(self.registry),
            {
                FileUploadService,
                GravatarService,
                URLAvatarService,
            })
        self.assertEqual(
            {
                avatar_service.avatar_service_id
                for avatar_service in self.registry.enabled_services
            },
            set(expected_enabled_services))

        # Verify that the saved settings reflect the correct state. These
        # checks include the defaults:
        siteconfig = SiteConfiguration.objects.get(pk=self.siteconfig.pk)
        self.assertIs(siteconfig.get(AVATARS_ENABLED_KEY), expect_enabled)
        self.assertEqual(siteconfig.get(ENABLED_SERVICES_KEY),
                         expected_enabled_services)
        self.assertEqual(siteconfig.get(DEFAULT_SERVICE_KEY),
                         expected_default_service)
        self.assertIsNone(siteconfig.get(LEGACY_AVATARS_MIGRATED_KEY))
        self.assertIsNone(siteconfig.get(LEGACY_INTEGRATION_GRAVATARS_KEY))

        # These checks are for what's actually written to siteconfig.
        siteconfig_settings = siteconfig.settings
        self.assertNotIn(LEGACY_AVATARS_MIGRATED_KEY, siteconfig_settings)
        self.assertNotIn(LEGACY_INTEGRATION_GRAVATARS_KEY, siteconfig_settings)

        if expect_has_enabled_saved:
            self.assertIs(siteconfig_settings.get(AVATARS_ENABLED_KEY),
                          expect_enabled)
        else:
            self.assertNotIn(AVATARS_ENABLED_KEY, siteconfig_settings)

        if expect_has_services_saved:
            self.assertIn(ENABLED_SERVICES_KEY, siteconfig_settings)
            self.assertIn(DEFAULT_SERVICE_KEY, siteconfig_settings)
            self.assertEqual(siteconfig_settings[ENABLED_SERVICES_KEY],
                             expected_enabled_services)
            self.assertEqual(siteconfig_settings[DEFAULT_SERVICE_KEY],
                             expected_default_service)
        else:
            self.assertNotIn(ENABLED_SERVICES_KEY, siteconfig_settings)
            self.assertNotIn(DEFAULT_SERVICE_KEY, siteconfig_settings)


class TemplateTagTests(AvatarServicesTestMixin, TestCase):
    """Tests for reviewboard.avatars.templatetags."""

    fixtures = ['test_users']

    def setUp(self):
        super(TemplateTagTests, self).setUp()

        avatar_services.enable_service(GravatarService, save=False)
        avatar_services.set_default_service(GravatarService, save=False)

        self.user = User.objects.get(username='doc')
        self.request = RequestFactory().get('/')
        self.request.user = AnonymousUser()

    def test_avatar_urls(self):
        """Testing {% avatar_urls %} template tag"""
        service = avatar_services.default_service
        self.assertIsNotNone(service)

        t = Template(
            '{% load avatars %}'
            '{% avatar_urls u 32 %}'
        )

        self.assertEqual(
            t.render(RequestContext(self.request, {
                'u': self.user,
                'service_id': service.avatar_service_id,
            })),
            ('{'
             '"1x": "%(1x)s", '
             '"2x": "%(2x)s", '
             '"3x": "%(3x)s"'
             '}'
             % service.get_avatar_urls_uncached(self.user, 32))
        )

    def test_avatar_urls_with_service(self):
        """Testing {% avatar_urls %} template tag with avatar_service_id"""
        avatar_services.register(DummyHighDPIAvatarService)
        avatar_services.enable_service(DummyHighDPIAvatarService)

        service = avatar_services.get_avatar_service(
            DummyHighDPIAvatarService.avatar_service_id)

        t = Template(
            '{% load avatars %}'
            '{% avatar_urls u 32 service_id %}'
        )

        self.assertEqual(
            t.render(RequestContext(self.request, {
                'u': self.user,
                'service_id': DummyHighDPIAvatarService.avatar_service_id,
            })),
            ('{'
             '"1x": "%(1x)s", '
             '"2x": "%(2x)s"'
             '}'
             % service.get_avatar_urls_uncached(self.user, 32))
        )

    def test_avatar_urls_no_service(self):
        """Testing {% avatar_urls %} template tag with no available services"""
        services = list(avatar_services)

        for service in services:
            avatar_services.unregister(service)

        t = Template(
            '{% load avatars %}'
            '{% avatar_urls u 32 %}'
        )

        self.assertEqual(
            t.render(RequestContext(self.request, {
                'u': self.user,
            })),
            '{}')

    def test_avatar_urls_service_not_found(self):
        """Testing {% avatar_urls %} template tag with an invalid service"""
        service = avatar_services.default_service

        self.assertIsNotNone(service)
        self.assertIsNone(avatar_services.get_avatar_service(
            DummyAvatarService.avatar_service_id))

        t = Template(
            '{% load avatars %}'
            '{% avatar_urls u 32 service_id %}'
        )

        self.assertEqual(
            t.render(RequestContext(self.request, {
                'u': self.user,
                'service_id': DummyAvatarService.avatar_service_id,
            })),
            ('{'
             '"1x": "%(1x)s", '
             '"2x": "%(2x)s", '
             '"3x": "%(3x)s"'
             '}'
             % service.get_avatar_urls_uncached(self.user, 32))
        )

    def test_avatar_default_service(self):
        """Testing {% avatar %} template tag with the default avatar service"""
        default_avatar_template = Template('{% load avatars %}'
                                           '{% avatar target_user 32 %}')
        service_template = Template(
            '{% load avatars %}'
            '{% avatar target_user 32 avatar_service_id %}')

        self.assertIsNotNone(avatar_services.default_service)

        self.assertHTMLEqual(
            default_avatar_template.render(RequestContext(self.request, {
                'target_user': self.user,
            })),
            service_template.render(RequestContext(self.request, {
                'target_user': self.user,
                'avatar_service_id': GravatarService.avatar_service_id,
            })))

    def test_avatar_specific_service(self):
        """Testing {% avatar %} template tag using a specific avatar service"""
        avatar_services.register(DummyAvatarService)
        avatar_services.enable_service(DummyAvatarService)

        t = Template('{% load avatars %}'
                     '{% avatar target_user 32 avatar_service_id %}')

        self.assertHTMLEqual(
            t.render(RequestContext(self.request, {
                'target_user': self.user,
                'avatar_service_id': DummyAvatarService.avatar_service_id,

            })),
            ('<img src="http://example.com/avatar.png" alt="%s" width="32"'
             ' height="32" srcset="http://example.com/avatar.png 1x"'
             ' class="avatar">\n'
             % self.user.username)
        )

    def test_avatar_invalid_service(self):
        """Test {% avatar %} template tag rendering with an invalid avatar
        service
        """
        t = Template('{% load avatars %}'
                     '{% avatar target_user 32 avatar_service_id %}')

        self.assertEqual(
            t.render(RequestContext(self.request, {
                'target_user': self.user,
                'avatar_service_id': 'INVALID_ID',
            })),
            t.render(RequestContext(self.request, {
                'target_user': self.user,
                'avatar_service_id': None,
            })))

    def test_username_unsafe(self):
        """Testing avatar template tag rendering with an unsafe username"""
        t = Template('{% load avatars %}'
                     '{% avatar target_user 32 avatar_service_id %}')

        self.request.user = User.objects.get(username='admin')

        user = User.objects.create_user(
            first_name='<b>Bad',
            last_name='User</b>',
            username='bad_user',
            email='bad_user@example.com')

        Profile.objects.create(user=user)

        avatar_services.register(DummyAvatarService)
        avatar_services.enable_service(DummyAvatarService)

        self.assertHTMLEqual(
            t.render(RequestContext(self.request, {
                'target_user': user,
                'avatar_service_id': DummyAvatarService.avatar_service_id,
            })),
            '<img src="http://example.com/avatar.png" alt="%s" width="32"'
            ' height="32" srcset="http://example.com/avatar.png 1x"'
            ' class="avatar">\n'
            % '&lt;b&gt;Bad User&lt;/b&gt;')


class FileUploadServiceTests(SpyAgency, AvatarServicesTestMixin, TestCase):
    fixtures = ['test_users']

    @classmethod
    def setUpClass(cls):
        super(FileUploadServiceTests, cls).setUpClass()

        cls.request_factory = RequestFactory()

    def test_absolute_urls(self):
        """Testing FileUploadService.get_avatar_urls_uncached returns absolute
        URLs
        """
        user = User.objects.get(username='doc')
        avatar = SimpleUploadedFile('filename.png', content=b' ',
                                    content_type='image/png')

        service = avatar_services.get_avatar_service(
            FileUploadService.avatar_service_id)
        storage_cls = get_storage_class()

        self.spy_on(storage_cls.save,
                    call_fake=lambda self, filename, data: filename)

        form = AvatarSettingsForm(
            None,
            self.request_factory.post('/'),
            user,
            data={
                'avatar_service_id': FileUploadService.avatar_service_id,
            })

        service_form = \
            form.avatar_service_forms[FileUploadService.avatar_service_id]

        form.files = service_form.files = {
            service_form.add_prefix('avatar_upload'): avatar,
        }

        self.assertTrue(form.is_valid())
        form.save()

        file_path = storage_cls.save.spy.last_call.args[0]

        self.assertEqual(service.get_avatar_urls_uncached(user, None),
                         {'1x': 'http://example.com/media/%s' % file_path})



from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _

from reviewboard.reviews.actions import (BaseReviewRequestAction,
                                         BaseReviewRequestMenuAction)
from reviewboard.reviews.features import general_comments_feature
from reviewboard.reviews.models import ReviewRequest
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.urls import diffviewer_url_names


class CloseMenuAction(BaseReviewRequestMenuAction):
    """A menu action for closing the corresponding review request."""

    action_id = 'close-review-request-action'
    label = _('Close')

    def should_render(self, context):
        review_request = context['review_request']

        return (review_request.status == ReviewRequest.PENDING_REVIEW and
                (context['request'].user.pk == review_request.submitter_id or
                 (context['perms']['reviews']['can_change_status'] and
                  review_request.public)))


class SubmitAction(BaseReviewRequestAction):
    """An action for submitting the review request."""

    action_id = 'submit-review-request-action'
    label = _('Submitted')

    def should_render(self, context):
        return context['review_request'].public


class DiscardAction(BaseReviewRequestAction):
    """An action for discarding the review request."""

    action_id = 'discard-review-request-action'
    label = _('Discarded')


class DeleteAction(BaseReviewRequestAction):
    """An action for permanently deleting the review request."""

    action_id = 'delete-review-request-action'
    label = _('Delete Permanently')

    def should_render(self, context):
        return context['perms']['reviews']['delete_reviewrequest']


class UpdateMenuAction(BaseReviewRequestMenuAction):
    """A menu action for updating the corresponding review request."""

    action_id = 'update-review-request-action'
    label = _('Update')

    def should_render(self, context):
        review_request = context['review_request']

        return (review_request.status == ReviewRequest.PENDING_REVIEW and
                (context['request'].user.pk == review_request.submitter_id or
                 context['perms']['reviews']['can_edit_reviewrequest']))


class UploadDiffAction(BaseReviewRequestAction):
    """An action for updating/uploading a diff for the review request."""

    action_id = 'upload-diff-action'

    def get_label(self, context):
        """Return this action's label.

        The label will change depending on whether or not the corresponding
        review request already has a diff.

        Args:
            context (django.template.Context):
                The collection of key-value pairs from the template.

        Returns:
            unicode: The label that displays this action to the user.
        """
        review_request = context['review_request']
        draft = review_request.get_draft(context['request'].user)

        if (draft and draft.diffset) or review_request.get_diffsets():
            return _('Update Diff')

        return _('Upload Diff')

    def should_render(self, context):
        """Return whether or not this action should render.

        If the corresponding review request has a repository, then an upload
        diff form exists, so we should render this UploadDiffAction.

        Args:
            context (django.template.Context):
                The collection of key-value pairs available in the template
                just before this action is to be rendered.

        Returns:
            bool: Determines if this action should render.
        """
        return context['review_request'].repository_id is not None


class UploadFileAction(BaseReviewRequestAction):
    """An action for uploading a file for the review request."""

    action_id = 'upload-file-action'
    label = _('Add File')


class DownloadDiffAction(BaseReviewRequestAction):
    """An action for downloading a diff from the review request."""

    action_id = 'download-diff-action'
    label = _('Download Diff')

    def get_url(self, context):
        """Return this action's URL.

        Args:
            context (django.template.Context):
                The collection of key-value pairs from the template.

        Returns:
            unicode: The URL to invoke if this action is clicked.
        """
        match = context['request'].resolver_match

        # We want to use a relative URL in the diff viewer as we will not be
        # re-rendering the page when switching between revisions.
        if match.url_name in diffviewer_url_names:
            return 'raw/'

        return local_site_reverse('raw-diff', context['request'], kwargs={
            'review_request_id': context['review_request'].display_id,
        })

    def get_hidden(self, context):
        """Return whether this action should be initially hidden to the user.

        Args:
            context (django.template.Context):
                The collection of key-value pairs from the template.

        Returns:
            bool: Whether this action should be initially hidden to the user.
        """
        match = context['request'].resolver_match

        if match.url_name in diffviewer_url_names:
            return match.url_name == 'view-interdiff'

        return super(DownloadDiffAction, self).get_hidden(context)

    def should_render(self, context):
        """Return whether or not this action should render.

        Args:
            context (django.template.Context):
                The collection of key-value pairs available in the template
                just before this action is to be rendered.

        Returns:
            bool: Determines if this action should render.
        """
        review_request = context['review_request']
        request = context['request']
        match = request.resolver_match

        # If we're on a diff viewer page, then this DownloadDiffAction should
        # initially be rendered, but possibly hidden.
        if match.url_name in diffviewer_url_names:
            return True

        return review_request.repository_id is not None


class EditReviewAction(BaseReviewRequestAction):
    """An action for editing a review intended for the review request."""

    action_id = 'review-action'
    label = _('Review')

    def should_render(self, context):
        return context['request'].user.is_authenticated()


class AddGeneralCommentAction(BaseReviewRequestAction):
    """An action for adding a new general comment to a review."""

    action_id = 'general-comment-action'
    label = _('Add General Comment')

    def should_render(self, context):
        request = context['request']
        return (request.user.is_authenticated() and
                general_comments_feature.is_enabled(request=request))


class ShipItAction(BaseReviewRequestAction):
    """An action for quickly approving the review request without comments."""

    action_id = 'ship-it-action'
    label = _('Ship It!')

    def should_render(self, context):
        return context['request'].user.is_authenticated()


def get_default_actions():
    """Return a copy of all the default actions.

    Returns:
        list of BaseReviewRequestAction: A copy of all the default actions.
    """
    return [
        CloseMenuAction([
            SubmitAction(),
            DiscardAction(),
            DeleteAction(),
        ]),
        UpdateMenuAction([
            UploadDiffAction(),
            UploadFileAction(),
        ]),
        DownloadDiffAction(),
        EditReviewAction(),
        AddGeneralCommentAction(),
        ShipItAction(),
    ]



from __future__ import unicode_literals

import json
import logging
from collections import defaultdict

from django import forms
from django.conf.urls import url
from django.http import HttpResponse, HttpResponseBadRequest
from django.template import RequestContext
from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.http import require_POST

from reviewboard.admin.server import build_server_url, get_server_url
from reviewboard.hostingsvcs.forms import HostingServiceForm
from reviewboard.hostingsvcs.hook_utils import (close_all_review_requests,
                                                get_repository_for_hook,
                                                get_review_request_id)
from reviewboard.hostingsvcs.service import HostingService
from reviewboard.site.urlresolvers import local_site_reverse


class GoogleCodeForm(HostingServiceForm):
    googlecode_project_name = forms.CharField(
        label=_('Project name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}))


class GoogleCodeHookViews(object):
    """Container class for hook views."""

    @staticmethod
    @require_POST
    def post_receive_hook_close_submitted(request, local_site_name=None,
                                          repository_id=None,
                                          hosting_service_id=None,
                                          hooks_uuid=None):
        """Close review requests as submitted automatically after a push.

        Args:
            request (django.http.HttpRequest):
                The request from the Bitbucket webhook.

            local_site_name (unicode):
                The local site name, if available.

            repository_id (int):
                The pk of the repository, if available.

            hosting_service_id (unicode):
                The name of the hosting service.

            hooks_uuid (unicode):
                The UUID of the configured webhook.

        Returns:
            django.http.HttpResponse:
            A response for the request.
        """
        repository = get_repository_for_hook(repository_id, hosting_service_id,
                                             local_site_name, hooks_uuid)

        try:
            payload = json.loads(request.body)
        except ValueError as e:
            logging.error('The payload is not in JSON format: %s', e,
                          exc_info=1)
            return HttpResponseBadRequest('Invalid payload format')

        server_url = get_server_url(request=request)
        review_request_id_to_commits_map = \
            GoogleCodeHookViews._get_review_request_id_to_commits_map(
                payload, server_url)

        if review_request_id_to_commits_map:
            close_all_review_requests(review_request_id_to_commits_map,
                                      local_site_name, repository,
                                      hosting_service_id)

        return HttpResponse()

    @staticmethod
    def _get_review_request_id_to_commits_map(payload, server_url):
        """Return a mapping of review request ID to a list of commits.

        Args:
            payload (dict):
                The decoded webhook payload.

            server_url (unicode):
                The URL of the Review Board server.

        Returns:
            dict:
            A mapping from review request ID to a list of matching commits from
            the payload.

        """
        # The Google Code payload is the same for SVN and Mercurial
        # repositories. There is no information in the payload as to
        # which SCM tool was used for the commit. That's why the only way
        # to close a review request through this hook is by adding the review
        # request id in the commit message.
        review_request_id_to_commits_map = defaultdict(list)
        branch_name = payload.get('repository_path')

        if not branch_name:
            return review_request_id_to_commits_map

        revisions = payload.get('revisions', [])

        for revision in revisions:
            revision_id = revision.get('revision')

            if len(revision_id) > 7:
                revision_id = revision_id[:7]

            commit_message = revision.get('message')
            review_request_id = get_review_request_id(commit_message,
                                                      server_url)
            review_request_id_to_commits_map[review_request_id].append(
                '%s (%s)' % (branch_name, revision_id))

        return review_request_id_to_commits_map


class GoogleCode(HostingService):
    name = 'Google Code'
    form = GoogleCodeForm
    supported_scmtools = ['Mercurial', 'Subversion']
    supports_repositories = True
    supports_bug_trackers = True
    has_repository_hook_instructions = True

    repository_url_patterns = [
        url(r'^hooks/(?P<hooks_uuid>[a-z0-9]+)/close-submitted/$',
            GoogleCodeHookViews.post_receive_hook_close_submitted,
            name='googlecode-hooks-close-submitted'),
    ]

    repository_fields = {
        'Mercurial': {
            'path': 'http://%(googlecode_project_name)s'
                    '.googlecode.com/hg',
            'mirror_path': 'https://%(googlecode_project_name)s'
                           '.googlecode.com/hg',
        },
        'Subversion': {
            'path': 'http://%(googlecode_project_name)s'
                    '.googlecode.com/svn',
            'mirror_path': 'https://%(googlecode_project_name)s'
                           '.googlecode.com/svn',
        },
    }
    bug_tracker_field = 'http://code.google.com/p/' \
                        '%(googlecode_project_name)s/' \
                        'issues/detail?id=%%s'

    def get_repository_hook_instructions(self, request, repository):
        """Returns instructions for setting up incoming webhooks."""
        webhook_endpoint_url = build_server_url(local_site_reverse(
            'googlecode-hooks-close-submitted',
            local_site=repository.local_site,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': repository.hosting_account.service_name,
                'hooks_uuid': repository.get_or_create_hooks_uuid(),
            }))
        add_webhook_url = (
            'https://code.google.com/p/%s/adminSource'
            % repository.extra_data['googlecode_project_name'])

        example_id = 123
        example_url = build_server_url(local_site_reverse(
            'review-request-detail',
            local_site=repository.local_site,
            kwargs={
                'review_request_id': example_id,
            }))

        return render_to_string(
            'hostingsvcs/googlecode/repo_hook_instructions.html',
            RequestContext(request, {
                'example_id': example_id,
                'example_url': example_url,
                'repository': repository,
                'server_url': get_server_url(),
                'add_webhook_url': add_webhook_url,
                'webhook_endpoint_url': webhook_endpoint_url,
            }))



"""Utilities for sending e-mail messages."""

from __future__ import unicode_literals

import logging

from django.contrib.auth.models import User
from django.db.models import Q
from djblets.mail.utils import (build_email_address,
                                build_email_address_for_user)

from reviewboard.accounts.models import ReviewRequestVisit
from reviewboard.changedescs.models import ChangeDescription
from reviewboard.reviews.models import Group


def build_recipients(user, review_request, extra_recipients=None,
                     limit_recipients_to=None):
    """Build the recipient sets for an e-mail.

    By default, the user sending the e-mail, the review request submitter (if
    they are active), all active reviewers, and all active members of review
    groups will be recipients of the e-mail.

    If the ``limit_recipients_to`` parameter is provided, the given ``user``
    and the review request submitter (if active) will still be recipients of
    the e-mail, but all reviewers and members of review groups will not.
    Instead, the recipients given in ``limit_recipients_to`` will be used.

    Args:
        user (django.contrib.auth.models.User):
            The user sending the e-mail.

        review_request (reviewboard.reviews.models.ReviewRequest):
            The review request the e-mail corresponds to.

        extra_recipients (list):
            An optional list of extra recipients as
            :py:class:`Users <django.contrib.auth.models.User>` and
            :py:class:`Groups <reviewboard.reviews.models.Group>` that will
            receive the e-mail.

        limit_recipients_to (list):
            An optional list of recipients as
            :py:class:`Users <django.contrib.auth.models.User>` and
            :py:class:`Groups <reviewboard.reviews.models.Group>` who will
            receive the e-mail in place of the normal recipients.

    Returns:
        tuple:
        A 2-tuple of the To field and the CC field, as sets of
        :py:class:`Users <django.contrib.auth.models.User>` and
        :py:class:`Groups <reviewboard.reviews.models.Group>`.
    """
    recipients = set()
    to_field = set()

    local_site = review_request.local_site_id
    submitter = review_request.submitter

    target_people = review_request.target_people.filter(is_active=True).extra(
        select={
            'visibility': """
                SELECT accounts_reviewrequestvisit.visibility
                  FROM accounts_reviewrequestvisit
                 WHERE accounts_reviewrequestvisit.review_request_id =
                       reviews_reviewrequest_target_people.reviewrequest_id
                   AND accounts_reviewrequestvisit.user_id =
                       reviews_reviewrequest_target_people.user_id
            """
        })

    starred_users = User.objects.filter(
        is_active=True,
        profile__starred_review_requests=review_request,
        profile__should_send_email=True)

    local_site_q = Q()

    if local_site:
        # Filter out users who are on the reviewer list in some form or have
        # starred the review request but are no longer part of the LocalSite.
        local_site_q = (Q(local_site=local_site) |
                        Q(local_site_admins=local_site))

        target_people = target_people.filter(local_site_q)

        starred_users = starred_users.filter(local_site_q)

    if not extra_recipients:
        extra_recipients = User.objects.none()

    if user.should_send_email():
        recipients.add(user)

    try:
        changedesc = review_request.changedescs.latest()
    except ChangeDescription.DoesNotExist:
        pass
    else:
        # If the submitter has changed and the person sending this e-mail is
        # not the original submitter, then we should include them in the list
        # of recipients.
        if changedesc.fields_changed:
            submitter_info = changedesc.fields_changed.get('submitter')

            if submitter_info:
                prev_submitter_pk = submitter_info['old'][0][2]
                prev_submitter = User.objects.get(pk=prev_submitter_pk)

                if (prev_submitter.is_active and
                    prev_submitter.should_send_email()):
                    recipients.add(prev_submitter)

    if submitter.is_active and submitter.should_send_email():
        recipients.add(submitter)

    recipients.update(starred_users)

    def _filter_recipients(to_filter):
        """Filter the given recipients.

        All groups will be added to the resulting recipients. Only users with a
        matching local site will be added to the resulting recipients.

        Args:
            to_filter (list):
                A list of recipients as
                :py:class:`Users <django.contrib.auth.models.User>` and
                :py:class:`Groups <reviewboard.reviews.models.Group>`.
        """
        pks = set()

        for recipient in to_filter:
            if isinstance(recipient, User):
                pks.add(recipient.pk)
            elif isinstance(recipient, Group):
                recipients.add(recipient)
            else:
                logging.error(
                    'Unexpected e-mail recipient %r; expected '
                    'django.contrib.auth.models.User or '
                    'reviewboard.reviews.models.Group.',
                    recipient)
        if pks:
            filtered_users = User.objects.filter(
                Q(is_active=True, pk__in=pks),
                local_site_q)

            recipients.update(
                recipient
                for recipient in filtered_users.select_related('profile')
                if recipient.should_send_email()
            )

    if limit_recipients_to is not None:
        _filter_recipients(limit_recipients_to)
    else:
        _filter_recipients(extra_recipients)

        to_field.update(
            recipient
            for recipient in target_people.select_related('profile')
            if (recipient.should_send_email() and
                recipient.visibility != ReviewRequestVisit.MUTED)
        )

        recipients.update(to_field)
        recipients.update(review_request.target_groups.all())

    if not user.should_send_own_updates():
        recipients.discard(user)
        to_field.discard(user)

    if to_field:
        cc_field = recipients.symmetric_difference(to_field)
    else:
        to_field = recipients
        cc_field = set()

    return to_field, cc_field


def get_email_addresses_for_group(group, review_request_id=None):
    """Build a list of e-mail addresses for the group.

    Args:
        group (reviewboard.reviews.models.Group):
            The review group to build the e-mail addresses for.

        review_request_id (int, optional):


    Returns:
        list of unicode:
        A list of properly formatted e-mail addresses for all users in the
        review group.
    """
    addresses = []

    if group.mailing_list:
        if ',' not in group.mailing_list:
            # The mailing list field has only one e-mail address in it,
            # so we can just use that and the group's display name.
            addresses = [build_email_address(full_name=group.display_name,
                                             email=group.mailing_list)]
        else:
            # The mailing list field has multiple e-mail addresses in it.
            # We don't know which one should have the group's display name
            # attached to it, so just return their custom list as-is.
            addresses = group.mailing_list.split(',')

    if not (group.mailing_list and group.email_list_only):
        users_q = Q(is_active=True)

        local_site = group.local_site

        if local_site:
            users_q = users_q & (Q(local_site=local_site) |
                                 Q(local_site_admins=local_site))

        users = group.users.filter(users_q).select_related('profile')

        if review_request_id:
            users = users.extra(select={
                'visibility': """
                    SELECT accounts_reviewrequestvisit.visibility
                      FROM accounts_reviewrequestvisit
                     WHERE accounts_reviewrequestvisit.review_request_id =
                           %s
                       AND accounts_reviewrequestvisit.user_id =
                           reviews_group_users.user_id
                """ % review_request_id
            })

        addresses.extend([
            build_email_address_for_user(u)
            for u in users
            if (u.should_send_email() and
                (not review_request_id or
                 u.visibility != ReviewRequestVisit.MUTED))
        ])

    return addresses


def recipients_to_addresses(recipients, review_request_id=None):
    """Return the set of e-mail addresses for the recipients.

    Args:
        recipients (list):
            A list of :py:class:`Users <django.contrib.auth.models.User>` and
            :py:class:`Groups <reviewboard.reviews.models.Group>`.

    Returns:
        set: The e-mail addresses for all recipients.
    """
    addresses = set()

    for recipient in recipients:
        assert isinstance(recipient, User) or isinstance(recipient, Group)

        if isinstance(recipient, User):
            addresses.add(build_email_address_for_user(recipient))
        else:
            addresses.update(get_email_addresses_for_group(recipient,
                                                           review_request_id))

    return addresses


def send_email(email_builder, **kwargs):
    """Attempt to send an e-mail, logging any exceptions that occur.

    Args:
        email_builder (callable):
            A function that generates an :py:class:`EmailMessage`.

        **kwargs (dict):
            Keyword arguments to provide to ``email_builder``.

    Returns:
        tuple:
        A tuple of:

        * The message that was generated (:py:class`EmailMessage`).
        * Whether or not the message was sent successfully (:py:class:`bool`).
    """
    message = email_builder(**kwargs)

    if message is None:
        return None, False

    try:
        message.send()
    except Exception:
        logging.exception(
            'Could not send e-mail message with subject "%s" from "%s" to '
            '"%s"',
            message.subject,
            message.from_email,
            message.to + (message.cc or []))

        return message, False

    return message, True



from __future__ import unicode_literals

from django import forms
from django.utils.translation import ugettext_lazy as _

from reviewboard.admin.server import get_hostname
from reviewboard.hostingsvcs.forms import HostingServiceForm
from reviewboard.hostingsvcs.service import HostingService
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)


class AssemblaForm(HostingServiceForm):
    assembla_project_id = forms.CharField(
        label=_('Project ID'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_(
            "The project ID, as shown in the URL "
            "(https://www.assembla.com/spaces/<b>&lt;project_id&gt;</b>), or "
            "your Perforce repository's Depot Host."))

    def save(self, repository):
        """Save the Assembla repository form.

        This will force the Perforce host and ticket authentication settings
        to values required for Assembla.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository being saved.
        """
        super(AssemblaForm, self).save(repository)

        if repository.get_scmtool().name == 'Perforce':
            project_id = self.cleaned_data['assembla_project_id']
            client_name = Assembla.make_p4_client_name(project_id)

            repository.extra_data.update({
                'use_ticket_auth': True,
                'p4_host': project_id,
                'p4_client': client_name,
            })


class Assembla(HostingService):
    """Hosting service support for Assembla.com.

    Assembla is a hosting service that offers, amongst other features,
    Perforce, Subversion, and Git repository support.

    They do not have much of an API that we can take advantage of, so it's
    impossible for us to support Git. However, Perforce and Subversion work.
    """

    name = 'Assembla'

    needs_authorization = True
    supports_bug_trackers = True
    supports_repositories = True
    supported_scmtools = ['Perforce', 'Subversion']

    form = AssemblaForm

    repository_fields = {
        'Perforce': {
            'path': 'perforce.assembla.com:1666',
        },
        'Subversion': {
            'path': 'https://subversion.assembla.com/svn/'
                    '%(assembla_project_id)s/',
        },
    }

    bug_tracker_field = (
        'https://www.assembla.com/spaces/%(assembla_project_id)s/'
        'tickets/%%s'
    )

    @classmethod
    def make_p4_client_name(cls, project_id):
        """Return a new P4CLIENT value from the hostname and project ID.

        The client name will consist of the Review Board server's hostname
        and a sanitized version of the project ID.

        Args:
            project_id (unicode):
                The project ID provided by Assembla. This is equivalent to the
                P4HOST value for Perforce.

        Returns:
            unicode:
            A new Perforce client name.
        """
        return '%s-%s' % (get_hostname(), project_id.replace('/', '-'))

    def check_repository(self, path, username, password, scmtool_class,
                         local_site_name, assembla_project_id=None,
                         *args, **kwargs):
        """Check the validity of a repository hosted on Assembla.

        Perforce repositories are handled specially. The Assembla project ID
        will be used as the Perforce host, which is needed to tell Assembla
        which repository on the server to use.

        Args:
            path (unicode):
                The repository path.

            username (unicode):
                The username used for authenticating.

            password (unicode):
                The password used for authenticating.

            scmtool_class (type):
                The SCMTool for the repository.

            local_site_name (unicode):
                The name of the Local Site, if any.

            assembla_project_id (unicode):
                The project ID for the Assembla team.

            *args (tuple):
                Additional arguments to pass to the superclass.

            **kwargs (dict):
                Additional keyword arguments to pass to the superclass.
        """
        # We want to use the configured username and other information from
        # the account.
        username = self.account.username
        password = self.get_password()

        if scmtool_class.name == 'Perforce':
            scmtool_class.check_repository(
                path=path,
                username=username,
                password=password,
                local_site_name=local_site_name,
                p4_host=assembla_project_id,
                p4_client=self.make_p4_client_name(assembla_project_id))
        else:
            super(Assembla, self).check_repository(
                path=path,
                username=username,
                password=password,
                local_site_name=local_site_name,
                scmtool_class=scmtool_class,
                **kwargs)

    def authorize(self, username, password, *args, **kwargs):
        """Authorize the Assembla account.

        For Assembla, we simply use the native SCMTool support, as there's
        no useful API available. We just store the password encrypted, which
        will be used by the SCMTool.

        Args:
            username (unicode):
                The username for authentication.

            password (unicode):
                The password for authentication.

            *args (tuple):
                Additional arguments.

            **kwargs (dict):
                Additional keyword arguments.
        """
        self.account.data['password'] = encrypt_password(password)
        self.account.save()

    def is_authorized(self):
        """Return if the account has a password set.

        Returns:
            bool:
            ``True`` if a password is set, or ``False`` if one has not yet
            been set.
        """
        return self.account.data.get('password') is not None

    def get_password(self):
        """Return the password for this account.

        This is needed for Perforce and Subversion.

        Returns:
            unicode:
            The stored password for the account.
        """
        return decrypt_password(self.account.data['password'])

    @classmethod
    def get_repository_fields(cls, tool_name=None, *args, **kwargs):
        """Return values for the fields in the repository form.

        This forces the encoding value to "utf8" on Perforce, which is needed
        by Assembla.

        Args:
            tool_name (unicode):
                The name of the SCMTool for the repository.

            *args (tuple):
                Additional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict:
            The resulting repository field values.
        """
        data = super(Assembla, cls).get_repository_fields(tool_name=tool_name,
                                                          *args, **kwargs)

        if tool_name == 'Perforce':
            data['encoding'] = 'utf8'

        return data



# coding: utf-8
"""Unit tests for ReviewRequestUpdatesView."""

from __future__ import unicode_literals

import json
import struct
from datetime import datetime, timedelta

from django.core.urlresolvers import reverse
from django.utils.timezone import utc

from reviewboard.reviews.views import ReviewRequestUpdatesView
from reviewboard.testing import TestCase


class ReviewRequestUpdatesViewTests(TestCase):
    """Unit tests for ReviewRequestUpdatesView."""

    fixtures = ['test_users']

    def setUp(self):
        super(ReviewRequestUpdatesViewTests, self).setUp()

        self.review_request = self.create_review_request(
            publish=True,
            time_added=datetime(2017, 9, 7, 17, 0, 0, tzinfo=utc),
            last_updated=datetime(2017, 9, 7, 23, 10, 0, tzinfo=utc))

        # Create the first review.
        self.review1 = self.create_review(
            self.review_request,
            timestamp=self.review_request.time_added + timedelta(days=10),
            publish=True)
        self.general_comment = self.create_general_comment(
            self.review1,
            issue_opened=True)

        # Create the second review (10 days later).
        self.review2 = self.create_review(
            self.review_request,
            timestamp=self.review1.timestamp + timedelta(days=10),
            publish=True)

        # This one shouldn't appear in any updates, since it's not published.
        self.review3 = self.create_review(
            self.review_request,
            timestamp=self.review2.timestamp + timedelta(days=10),
            publish=False)

        self.view = ReviewRequestUpdatesView.as_view()

    def test_get(self):
        """Testing ReviewRequestUpdatesView GET"""
        updates = self._get_updates()
        self.assertEqual(len(updates), 4)

        metadata, html = updates[0]
        self.assertEqual(metadata['type'], 'entry')
        self.assertEqual(metadata['entryType'], 'review')
        self.assertEqual(metadata['entryID'], '1')
        self.assertEqual(metadata['addedTimestamp'],
                         '2017-09-17 17:00:00+00:00')
        self.assertEqual(metadata['updatedTimestamp'],
                         '2017-09-17 17:00:00+00:00')
        self.assertEqual(metadata['viewOptions'], {})
        self.assertEqual(metadata['modelData'], {
            'reviewData': {
                'id': self.review1.pk,
                'public': True,
                'bodyTop': self.review1.body_top,
                'bodyBottom': self.review1.body_bottom,
                'shipIt': self.review1.ship_it,
            },
        })
        self.assertTrue(html.startswith('<div id="review1"'))
        self.assertTrue(html.endswith('\n</div>'))

        metadata, html = updates[1]
        self.assertEqual(metadata['type'], 'entry')
        self.assertEqual(metadata['entryType'], 'review')
        self.assertEqual(metadata['entryID'], '2')
        self.assertEqual(metadata['addedTimestamp'],
                         '2017-09-27 17:00:00+00:00')
        self.assertEqual(metadata['updatedTimestamp'],
                         '2017-09-27 17:00:00+00:00')
        self.assertEqual(metadata['viewOptions'], {})
        self.assertEqual(metadata['modelData'], {
            'reviewData': {
                'id': self.review2.pk,
                'public': True,
                'bodyTop': self.review2.body_top,
                'bodyBottom': self.review2.body_bottom,
                'shipIt': self.review2.ship_it,
            },
        })
        self.assertTrue(html.startswith('<div id="review2"'))
        self.assertTrue(html.endswith('\n</div>'))

        metadata, html = updates[2]
        self.assertEqual(metadata['type'], 'entry')
        self.assertEqual(metadata['entryType'], 'initial_status_updates')
        self.assertEqual(metadata['entryID'], '0')
        self.assertEqual(metadata['addedTimestamp'],
                         '2017-09-07 17:00:00+00:00')
        self.assertEqual(metadata['updatedTimestamp'],
                         '2017-09-07 17:00:00+00:00')
        self.assertEqual(metadata['viewOptions'], {})
        self.assertEqual(metadata['modelData'], {
            'pendingStatusUpdates': False,
        })
        self.assertTrue(html.startswith('<div id="initial_status_updates"'))
        self.assertTrue(html.endswith('\n</div>'))

        metadata, html = updates[3]
        self.assertEqual(metadata['type'], 'issue-summary-table')
        self.assertTrue(html.startswith('<div id="issue-summary"'))
        self.assertTrue(html.endswith('\n</div>'))

    def test_get_with_unicode(self):
        """Testing ReviewRequestUpdatesView GET with Unicode content"""
        # Add some Unicode content.
        self.review1.body_top = 'áéíóú'
        self.review1.save(update_fields=('body_top',))

        self.review2.body_top = 'ÄËÏÖÜŸ'
        self.review2.save(update_fields=('body_top',))

        self.general_comment.text = 'ĀĒĪŌ'
        self.general_comment.save(update_fields=('text',))

        updates = self._get_updates()
        self.assertEqual(len(updates), 4)

        metadata, html = updates[0]
        self.assertEqual(metadata['type'], 'entry')
        self.assertEqual(metadata['entryType'], 'review')
        self.assertEqual(metadata['entryID'], '1')
        self.assertEqual(metadata['addedTimestamp'],
                         '2017-09-17 17:00:00+00:00')
        self.assertEqual(metadata['updatedTimestamp'],
                         '2017-09-17 17:00:00+00:00')
        self.assertEqual(metadata['viewOptions'], {})
        self.assertEqual(metadata['modelData'], {
            'reviewData': {
                'id': self.review1.pk,
                'public': True,
                'bodyTop': self.review1.body_top,
                'bodyBottom': self.review1.body_bottom,
                'shipIt': self.review1.ship_it,
            },
        })
        self.assertTrue(html.startswith('<div id="review1"'))
        self.assertTrue(html.endswith('\n</div>'))
        self.assertIn('áéíóú', html)

        metadata, html = updates[1]
        self.assertEqual(metadata['type'], 'entry')
        self.assertEqual(metadata['entryType'], 'review')
        self.assertEqual(metadata['entryID'], '2')
        self.assertEqual(metadata['addedTimestamp'],
                         '2017-09-27 17:00:00+00:00')
        self.assertEqual(metadata['updatedTimestamp'],
                         '2017-09-27 17:00:00+00:00')
        self.assertEqual(metadata['viewOptions'], {})
        self.assertEqual(metadata['modelData'], {
            'reviewData': {
                'id': self.review2.pk,
                'public': True,
                'bodyTop': self.review2.body_top,
                'bodyBottom': self.review2.body_bottom,
                'shipIt': self.review2.ship_it,
            },
        })
        self.assertTrue(html.startswith('<div id="review2"'))
        self.assertTrue(html.endswith('\n</div>'))
        self.assertIn('ÄËÏÖÜŸ', html)

        metadata, html = updates[2]
        self.assertEqual(metadata['type'], 'entry')
        self.assertEqual(metadata['entryType'], 'initial_status_updates')
        self.assertEqual(metadata['entryID'], '0')
        self.assertEqual(metadata['addedTimestamp'],
                         '2017-09-07 17:00:00+00:00')
        self.assertEqual(metadata['updatedTimestamp'],
                         '2017-09-07 17:00:00+00:00')
        self.assertEqual(metadata['viewOptions'], {})
        self.assertEqual(metadata['modelData'], {
            'pendingStatusUpdates': False,
        })
        self.assertTrue(html.startswith('<div id="initial_status_updates"'))
        self.assertTrue(html.endswith('\n</div>'))

        metadata, html = updates[3]
        self.assertEqual(metadata['type'], 'issue-summary-table')
        self.assertTrue(html.startswith('<div id="issue-summary"'))
        self.assertTrue(html.endswith('\n</div>'))
        self.assertIn('ĀĒĪŌ', html)

    def test_get_with_entries(self):
        """Testing ReviewRequestUpdatesView GET with ?entries=..."""
        updates = self._get_updates({
            'entries': 'review:2;initial_status_updates:0',
        })
        self.assertEqual(len(updates), 3)

        metadata, html = updates[0]
        self.assertEqual(metadata['type'], 'entry')
        self.assertEqual(metadata['entryType'], 'review')
        self.assertEqual(metadata['entryID'], '2')
        self.assertEqual(metadata['addedTimestamp'],
                         '2017-09-27 17:00:00+00:00')
        self.assertEqual(metadata['updatedTimestamp'],
                         '2017-09-27 17:00:00+00:00')
        self.assertEqual(metadata['viewOptions'], {})
        self.assertEqual(metadata['modelData'], {
            'reviewData': {
                'id': self.review2.pk,
                'public': True,
                'bodyTop': self.review2.body_top,
                'bodyBottom': self.review2.body_bottom,
                'shipIt': self.review2.ship_it,
            },
        })
        self.assertTrue(html.startswith('<div id="review2"'))
        self.assertTrue(html.endswith('\n</div>'))

        metadata, html = updates[1]
        self.assertEqual(metadata['type'], 'entry')
        self.assertEqual(metadata['entryType'], 'initial_status_updates')
        self.assertEqual(metadata['entryID'], '0')
        self.assertEqual(metadata['addedTimestamp'],
                         '2017-09-07 17:00:00+00:00')
        self.assertEqual(metadata['updatedTimestamp'],
                         '2017-09-07 17:00:00+00:00')
        self.assertEqual(metadata['viewOptions'], {})
        self.assertEqual(metadata['modelData'], {
            'pendingStatusUpdates': False,
        })
        self.assertTrue(html.startswith('<div id="initial_status_updates"'))
        self.assertTrue(html.endswith('\n</div>'))

        # The issue summary table is always added when reviews have updates.
        metadata, html = updates[2]
        self.assertEqual(metadata['type'], 'issue-summary-table')
        self.assertTrue(html.startswith('<div id="issue-summary"'))
        self.assertTrue(html.endswith('\n</div>'))

    def test_get_with_review_entries_adds_issue_summary_table(self):
        """Testing ReviewRequestUpdatesView GET with ?entries=review:...
        always includes the issue summary table
        """
        updates = self._get_updates({
            'entries': 'review:2',
        })
        self.assertEqual(len(updates), 2)

        metadata, html = updates[0]
        self.assertEqual(metadata['type'], 'entry')
        self.assertEqual(metadata['entryType'], 'review')
        self.assertEqual(metadata['entryID'], '2')
        self.assertEqual(metadata['addedTimestamp'],
                         '2017-09-27 17:00:00+00:00')
        self.assertEqual(metadata['updatedTimestamp'],
                         '2017-09-27 17:00:00+00:00')
        self.assertEqual(metadata['viewOptions'], {})
        self.assertEqual(metadata['modelData'], {
            'reviewData': {
                'id': self.review2.pk,
                'public': True,
                'bodyTop': self.review2.body_top,
                'bodyBottom': self.review2.body_bottom,
                'shipIt': self.review2.ship_it,
            },
        })
        self.assertTrue(html.startswith('<div id="review2"'))
        self.assertTrue(html.endswith('\n</div>'))

        metadata, html = updates[1]
        self.assertEqual(metadata['type'], 'issue-summary-table')
        self.assertTrue(html.startswith('<div id="issue-summary"'))
        self.assertTrue(html.endswith('\n</div>'))

    def test_get_with_invalid_entries(self):
        """Testing ReviewRequestUpdatesView GET with invalid ?entries=...
        values
        """
        response = self.client.get(self._build_url(), {
            'entries': 'review2',
        })
        self.assertEqual(response.status_code, 400)

    def test_get_with_since(self):
        """Testing ReviewRequestUpdatesView GET with ?since=..."""
        timestamp = self.review1.timestamp + timedelta(days=1)
        updates = self._get_updates({
            'since': timestamp.isoformat(),
        })
        self.assertEqual(len(updates), 2)

        metadata, html = updates[0]
        self.assertEqual(metadata['type'], 'entry')
        self.assertEqual(metadata['entryType'], 'review')
        self.assertEqual(metadata['entryID'], '2')
        self.assertEqual(metadata['addedTimestamp'],
                         '2017-09-27 17:00:00+00:00')
        self.assertEqual(metadata['updatedTimestamp'],
                         '2017-09-27 17:00:00+00:00')
        self.assertEqual(metadata['viewOptions'], {})
        self.assertEqual(metadata['modelData'], {
            'reviewData': {
                'id': self.review2.pk,
                'public': True,
                'bodyTop': self.review2.body_top,
                'bodyBottom': self.review2.body_bottom,
                'shipIt': self.review2.ship_it,
            },
        })
        self.assertTrue(html.startswith('<div id="review2"'))
        self.assertTrue(html.endswith('\n</div>'))

        metadata, html = updates[1]
        self.assertEqual(metadata['type'], 'issue-summary-table')
        self.assertTrue(html.startswith('<div id="issue-summary"'))
        self.assertTrue(html.endswith('\n</div>'))

    def test_post(self):
        """Testing ReviewRequestUpdatesView POST not allowed"""
        # 1 SQL query for SiteConfiguration in the middleware.
        with self.assertNumQueries(1):
            response = self.client.post(self._build_url())

        self.assertEqual(response.status_code, 405)

    def test_put(self):
        """Testing ReviewRequestUpdatesView PUT not allowed"""
        # 1 SQL query for SiteConfiguration in the middleware.
        with self.assertNumQueries(1):
            response = self.client.put(self._build_url())

        self.assertEqual(response.status_code, 405)

    def test_delete(self):
        """Testing ReviewRequestUpdatesView DELETE not allowed"""
        # 1 SQL query for SiteConfiguration in the middleware.
        with self.assertNumQueries(1):
            response = self.client.delete(self._build_url())

        self.assertEqual(response.status_code, 405)

    def _build_url(self):
        return reverse('review-request-updates',
                       args=[self.review_request.display_id])

    def _get_updates(self, query={}):
        response = self.client.get(self._build_url(), query)
        self.assertEqual(response.status_code, 200)

        content = response.content
        self.assertIs(type(content), bytes)

        i = 0
        updates = []

        while i < len(content):
            # Read the length of the metadata.
            metadata_len = struct.unpack_from('<L', content, i)[0]
            i += 4

            # Read the metadata.
            metadata = json.loads(content[i:i + metadata_len].decode('utf-8'))
            i += metadata_len

            # Read the length of the HTML.
            html_len = struct.unpack_from('<L', content, i)[0]
            i += 4

            # Read the HTML.
            html = content[i:i + html_len].decode('utf-8')
            i += html_len

            updates.append((metadata, html))

        return updates



"""Sidebar item management for datagrids."""

from __future__ import unicode_literals

from django.template.context import RequestContext
from django.template.loader import render_to_string
from django.utils import six
from django.utils.six.moves.urllib.parse import urlencode

from reviewboard.site.urlresolvers import local_site_reverse


class BaseSidebarItem(object):
    """Base class for an item on the sidebar of a datagrid.

    Items can optionally have labels and counts associated with them.
    Depending on the subclass, it may also be able to nest items.

    They may also have custom templates, for more advanced rendering.

    See :py:class:`SidebarNavItem` and :py:class:`BaseSidebarSection` for the
    common types of sidebar items.

    Attributes:
        datagrid (djblets.datagrid.grids.DataGrid):
            The datagrid containing this item.

        sidebar (Sidebar):
            The sidebar containing this item.
    """

    #: The template to use for rendering this item in the sidebar.
    template_name = None

    #: The displayed label for the item.
    label = None

    #: The name of the optional CSS icon to use beside the label.
    icon_name = None

    #: The datagrid "view" to link to when clicking this item.
    #:
    #: This corresponds to the ``?view=`` parameter passed to the datagrid
    #: page.
    view_id = None

    #: Additional key/values to pass to the URL when clicking this item.
    #:
    #: If provided, this must be a dictionary of keys and values for the
    #: URL. The keys and values will be automatically URL-encoded.
    view_args = None

    #: Additional CSS classes to include for the item.
    css_classes = None

    def __init__(self, sidebar, datagrid):
        """Initialize the sidebar item.

        Args:
            sidebar (Sidebar):
                The sidebar containing this item.

            datagrid (djblets.datagrid.grids.DataGrid):
                The datagrid containing this item.
        """
        self.sidebar = sidebar
        self.datagrid = datagrid

    def get_url(self):
        """Return the URL used when clicking the item.

        By default, this builds a URL to the parent datagrid using
        the ``view_id`` and ``view_args`` attributes. If they are not
        set, then the item won't be clickable.

        Returns:
            unicode:
            The URL to the dashboard view represented by this item.
        """
        if not self.view_id and not self.view_args:
            return None

        if self.view_args:
            url_args = self.view_args.copy()
        else:
            url_args = {}

        if self.view_id:
            url_args['view'] = self.view_id

        return '%s?%s' % (self.datagrid.request.path, urlencode(url_args))

    def get_count(self):
        """Return the count shown for this item.

        By default, this shows nothing. Subclasses can override to display
        a count.

        Returns:
            int:
            The count to display beside the item, or ``None`` if no count
            should be displayed.
        """
        return None

    def is_visible(self):
        """Return whether the item is visible.

        By default, an item is visible. Subclasses can override this to
        control visibility.

        Returns:
            bool:
            ``True`` if the item is visible. ``False`` if it's hidden.
        """
        return True

    def is_active(self):
        """Return whether the item is currently active.

        The item will be active if the current page matches the URL
        associated with the item.

        Returns:
            bool:
            ``True`` if the item represents the active page. ``False`` if it
            does not.
        """
        if self.view_id is None:
            return False

        request = self.datagrid.request
        view_id = request.GET.get('view', self.sidebar.default_view_id)

        if view_id != self.view_id:
            return False

        if self.view_args:
            for key, value in six.iteritems(self.view_args):
                if request.GET.get(key) != value:
                    return False

        return True

    def render(self):
        """Render the item.

        Returns:
            django.utils.safestring.SafeText:
            The rendered HTML for the item.
        """
        count = self.get_count()
        context = {
            'datagrid': self.datagrid,
            'label': self.label,
            'icon_name': self.icon_name or '',
            'view_id': self.view_id,
            'view_args': self.view_args,
            'count': count,
            'has_count': count is not None,
            'url': self.get_url(),
            'active': self.is_active(),
            'css_classes': self.css_classes or [],
        }
        context.update(self.get_extra_context())

        return render_to_string(self.template_name,
                                RequestContext(self.datagrid.request, context))

    def get_extra_context(self):
        """Return extra context for the render.

        Returns:
            dict:
            A dictionary of additional template context. By default, this is
            empty.
        """
        return {}


class BaseSidebarSection(BaseSidebarItem):
    """Base class for a section of items on the sidebar.

    Subclasses can override this to define a section and provide items
    listed in the section.

    Sections can optionally be clickable and display a count.
    """

    template_name = 'datagrids/sidebar_section.html'

    def __init__(self, *args, **kwargs):
        """Initialize the section.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent class.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        super(BaseSidebarSection, self).__init__(*args, **kwargs)

        self.items = list(self.get_items())

    def get_items(self):
        """Return the items displayed in this section.

        Subclasses must override this and return or yield the items
        to be displayed.

        Returns:
            list of BaseSidebarItem:
            The list of items to display in the section.
        """
        raise NotImplementedError

    def is_visible(self):
        """Return whether the section is visible.

        By default, a section is visible if it has any item classes
        registered. Subclasses can override this to provide more specific
        logic.

        Returns:
            bool:
            ``True`` if the section is visible. ``False`` if it's hidden.
        """
        return len(self.items) > 0

    def get_extra_context(self):
        """Return extra context for the section.

        Subclasses that override this method must call the parent method.

        Returns:
            dict:
            Additional template context for the rendering of the section.
        """
        return {
            'items': self.items,
        }


class SidebarNavItem(BaseSidebarItem):
    """A typical navigation link item on the sidebar.

    This is the standard type of item added to sections on a sidebar. An item
    can contain an explicit URL or a resolvable URL name to link to. If not
    provided, the current datagrid page's URL will be used along with query
    arguments built from :py:attr:`view_id` and :py:attr:`view_args`.
    """

    template_name = 'datagrids/sidebar_nav_item.html'

    def __init__(self, section, label, icon_name=None, view_id=None,
                 view_args=None, count=None, url=None, url_name=None,
                 css_classes=None):
        """Initialize the item.

        Args:
            section (BaseSidebarSection):
                The section that should contain this item.

            label (unicode):
                The displayed label for this item.

            icon_name (unicode, optional):
                The name of the optional CSS icon to display beside the label.

            view_id (unicode, optional):
                The ID of the optional datagrid view to display when clicking
                the item. See :py:attr:`BaseSidebarItem.view_id` for more
                information.

            view_args (unicode, optional)
                Keys/values to display in the URL when clicking the item. See
                :py:attr:`BaseSidebarItem.view_args` for more information.

            count (int, optional):
                The count to display beside the label.

            url (unicode, optional):
                The optional URL to navigate to when clicked.

            url_name (unicode, optional):
                The optional URL name to resolve and navigate to when clicked.

            css_classes (list of unicode, optional):
                Additional CSS classes to apply to the item.
        """
        super(SidebarNavItem, self).__init__(section.sidebar, section.datagrid)

        self.label = label
        self.icon_name = icon_name
        self.view_id = view_id
        self.view_args = view_args
        self.count = count
        self.css_classes = css_classes
        self.url = url
        self.url_name = url_name

    def get_url(self):
        """Return the URL for the item.

        If :py:attr:`url` is set, that URL will be returned directly.

        If :py:attr:`url_name` is set instead, it will be resolved relative
        to any Local Site that might be accessed and used as the URL. Note that
        the URL can't require any parameters.

        If not explicit URL or name is provided, the current page is used along
        with query parameters built from :py:attr:`view_id` and
        :py:attr:`view_args`.

        Returns:
            unicode:
            The URL to navigate to when clicked.
        """
        if self.url:
            return self.url
        elif self.url_name:
            return local_site_reverse(self.url_name,
                                      request=self.datagrid.request)
        else:
            return super(SidebarNavItem, self).get_url()

    def get_count(self):
        """Return the count provided in the constructor.

        Subclasses can override this if they need additional logic to compute
        a count.

        Returns:
            int:
            The count to display beside the label, or ``None`` if no count
            should be shown.
        """
        return self.count


class Sidebar(object):
    """Provides a sidebar for a datagrid.

    A sidebar can have several item classes added to it of various types.
    These will be instantiated and rendered when rendering the datagrid.
    """

    def __init__(self, item_classes, default_view_id=None, css_classes=[]):
        """Initialize the sidebar.

        Args:
            item_classes (list of type):
                The list of :py:class:`BaseSidebarItem` subclasses to include
                by default in the sidebar.

            default_view_id (unicode, optional):
                The default "view" of the datagrid to display. This corresponds
                to a registered :py:attr:`BaseSidebarItem.view_id`.

            css_classes (list of unicode):
                The list of additional CSS classes to apply to the sidebar.
        """
        self._item_classes = []
        self.css_classes = css_classes
        self.default_view_id = default_view_id

        for item_cls in item_classes:
            self.add_item(item_cls)

    def add_item(self, item_cls):
        """Add an item class to the sidebar.

        Args:
            item_cls (type):
                The item to add to the sidebar. This must be a subclass of
                :py:class:`BaseSidebarItem`.
        """
        self._item_classes.append(item_cls)

    def remove_item(self, item_cls):
        """Remove an item class from the sidebar.

        Args:
            item_cls (type):
                The item to remove from the sidebar. This must be a subclass
                of :py:class:`BaseSidebarItem`.
        """
        self._item_classes.remove(item_cls)

    def get_items(self, datagrid):
        """Instantiate and return all items on the sidebar.

        Args:
            datagrid (djblets.datagrid.grids.DataGrid):
                The datagrid instance to associate with each item.

        Returns:
            list of DataGridSidebarItem:
            The list of instantiated items.
        """
        return [
            item_cls(self, datagrid)
            for item_cls in self._item_classes
        ]


class DataGridSidebarMixin(object):
    """A mixin for datagrids using a sidebar.

    This is meant to be used along with :py:class:`Sidebar`. It will initialize
    the sidebar, providing instances of all the items for the template.
    """

    def load_extra_state(self, *args, **kwargs):
        """Compute any extra state for the sidebar.

        This will set :py:attr:`sidebar_items` on the datagrid to a list of
        instantiated items.

        Args:
            *args (tuple):
                Additional positional arguments passed to the method.

            **kwargs (dict):
                Additional keyword arguments passed to the method.

        Returns:
            object:
            The result from the parent method on the datagrid.
        """
        result = super(DataGridSidebarMixin, self).load_extra_state(
            *args, **kwargs)

        self.sidebar_items = self.sidebar.get_items(self)

        return result



"""Unit tests for additions to django.contrib.auth.models.User."""

from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser, User
from djblets.testing.decorators import add_fixtures

from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class UserTests(TestCase):
    """Unit tests for additions to django.contrib.auth.models.User."""

    fixtures = ['test_users']

    def test_is_profile_visible_with_public(self):
        """Testing User.is_profile_visible with public profiles"""
        user1 = User.objects.get(username='admin')
        user2 = User.objects.get(username='doc')

        self.assertTrue(user1.is_profile_visible(user2))

    def test_is_profile_visible_with_private(self):
        """Testing User.is_profile_visible with private profiles"""
        user1 = User.objects.get(username='admin')
        user2 = User.objects.get(username='doc')

        profile = user1.get_profile()
        profile.is_private = True
        profile.save()

        self.assertFalse(user1.is_profile_visible(user2))
        self.assertTrue(user1.is_profile_visible(user1))

        user2.is_staff = True
        self.assertTrue(user1.is_profile_visible(user2))

    def test_is_profile_visible_unauthenticated(self):
        """Testing User.is_profile_visible with an unauthenticated user"""
        user = User.objects.get(username='doc')

        self.assertFalse(user.is_profile_visible(AnonymousUser()))

    def test_is_profile_visible_no_user(self):
        """Testing User.is_profile_visible with no user"""
        user = User.objects.get(username='doc')

        self.assertFalse(user.is_profile_visible(None))

    def test_is_profile_visible_staff(self):
        """Testing User.is_profile_public with a staff user"""
        user = User.objects.get(username='doc')
        admin = User.objects.get(username='admin')

        profile = user.get_profile()
        profile.is_private = True
        profile.save()

        self.assertTrue(user.is_profile_visible(admin))

    def test_is_profile_visible_owner(self):
        """Testing User.is_profile_visible for the profile owner"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True
        profile.save()

        self.assertTrue(user.is_profile_visible(user))

    def test_is_profile_visible_local_site_member(self):
        """Testing User.is_profile_visible for a LocalSite member viewing a
        LocalSite member with a public profile
        """
        to_view = User.objects.get(username='doc')
        viewer = User.objects.get(username='grumpy')

        site = LocalSite.objects.create()
        site.users = [to_view, viewer]

        self.assertTrue(to_view.is_profile_visible(viewer))

    def test_is_profile_visible_local_site_member_private(self):
        """Testing User.is_profile_visible for a LocalSite member viewing a
        LocalSite member with a private profile
        """
        to_view = User.objects.get(username='doc')
        viewer = User.objects.get(username='grumpy')

        profile = to_view.get_profile()
        profile.is_private = True
        profile.save()

        site = LocalSite.objects.create()
        site.users = [to_view, viewer]

        self.assertFalse(to_view.is_profile_visible(viewer))

    def test_is_profile_visible_local_site_admin(self):
        """Testing user.is_profile_visible for a LocalSite admin viewing a
        LocalSite member with a public profile
        """
        to_view = User.objects.get(username='doc')
        viewer = User.objects.get(username='grumpy')

        site = LocalSite.objects.create()
        site.users = [to_view, viewer]
        site.admins = [viewer]

        self.assertTrue(to_view.is_profile_visible(viewer))

    def test_is_profile_visible_local_site_admin_private(self):
        """Testing user.is_profile_visible for a LocalSite admin viewing a
        LocalSite member with a private profile
        """
        to_view = User.objects.get(username='doc')
        viewer = User.objects.get(username='grumpy')

        profile = to_view.get_profile()
        profile.is_private = True
        profile.save()

        site = LocalSite.objects.create()
        site.users = [to_view, viewer]
        site.admins = [viewer]

        self.assertTrue(to_view.is_profile_visible(viewer))

    def test_is_admin_for_user_admin_vs_user(self):
        """Testing User.is_admin_for_user for an admin"""
        admin = User.objects.get(username='admin')
        user = User.objects.get(username='doc')

        with self.assertNumQueries(0):
            self.assertTrue(admin.is_admin_for_user(user))

    def test_is_admin_for_user_admin_vs_none(self):
        """Testing User.is_admin_for_user for an admin when the user is None"""
        admin = User.objects.get(username='admin')

        with self.assertNumQueries(0):
            self.assertTrue(admin.is_admin_for_user(None))

    def test_is_admin_for_user_admin_vs_anonymous(self):
        """Testing User.is_admin_for_user for an admin when the user is
        anonymous
        """
        admin = User.objects.get(username='admin')

        with self.assertNumQueries(0):
            self.assertTrue(admin.is_admin_for_user(AnonymousUser()))

    def test_is_admin_for_user_user_vs_user(self):
        """Testing User.is_admin_for_user for a regular user"""
        user = User.objects.get(username='doc')

        with self.assertNumQueries(1):
            self.assertFalse(user.is_admin_for_user(user))

        with self.assertNumQueries(0):
            self.assertFalse(user.is_admin_for_user(user))

    @add_fixtures(['test_site'])
    def test_is_admin_for_user_localsite_admin_vs_localsite_user(self):
        """Testing User.is_admin_for_user for a LocalSite admin when the user
        is a member of that LocalSite
        """
        site_admin = User.objects.get(username='doc')
        site_user = User.objects.get(username='admin')

        with self.assertNumQueries(1):
            self.assertTrue(site_admin.is_admin_for_user(site_user))

        with self.assertNumQueries(0):
            self.assertTrue(site_admin.is_admin_for_user(site_user))

    @add_fixtures(['test_site'])
    def test_is_admin_for_user_localsite_admin_vs_other_localsite_user(self):
        """Testing User.is_admin_for_user for a LocalSite admin when the user
        is a member of another LocalSite
        """
        site_admin = User.objects.get(username='doc')
        site_user = User.objects.get(username='grumpy')
        site = LocalSite.objects.create(name='local-site-3')
        site.users.add(site_admin)
        site.users.add(site_user)

        with self.assertNumQueries(1):
            self.assertFalse(site_admin.is_admin_for_user(site_user))

        with self.assertNumQueries(0):
            self.assertFalse(site_admin.is_admin_for_user(site_user))
